#include <sys/types.h>
#include <stddef.h>
#include <stdbool.h>

#include "scythe/object.h"
#include "scythe/earray.h"
#include "scythe/bedrock.h"
#include "scythe/vminst.h"
#include "scythe/vm.h"
#include "scythe/memory.h"
#include "scythe/refstk.h"
#include "scythe/number.h"
#include "scythe/compiler.h"
#include "scythe/equivalence.h"
#include "scythe/exception.h"
#include "scythe/iseq.h"
#include "scythe/miscobjects.h"
#include "scythe/module.h"
#include "scythe/pair.h"
#include "scythe/procedure.h"
#include "scythe/symbol.h"
#include "scythe/assembler.h"


static ScmObj
prog_get_mod_decls(ScmObj prog)
{
  scm_assert(scm_cdprogram_p(prog));
  return scm_cdprogram_log(prog);
}

static void
prog_set_mod_decls(ScmObj prog, ScmObj decls)
{
  scm_assert(scm_cdprogram_p(prog));
  scm_assert(scm_asmmoddecls_p(decls));
  scm_cdprogram_set_log(prog, decls);
}


/***************************************************************************/
/*  Assembler Instruction Format                                           */
/***************************************************************************/

static const int asm_inst_fmt_table[SCM_VM_INST_NR_OP] = {
    SCM_ASM_FMT_NOOPD,            /* SCM_OPCODE_NOP */
    SCM_ASM_FMT_NOOPD,            /* SCM_OPCODE_HALT */
    SCM_ASM_FMT_SI,               /* SCM_OPCODE_INT */
    SCM_ASM_FMT_IOF,              /* SCM_OPCODE_CFRAME */
    SCM_ASM_FMT_SI,               /* SCM_OPCODE_EFRAME */
    SCM_ASM_FMT_NOOPD,            /* SCM_OPCODE_EPOP */
    SCM_ASM_FMT_SI   ,            /* SCM_OPCODE_ESHIFT */
    SCM_ASM_FMT_OBJ,              /* SCM_OPCODE_IMMVAL */
    SCM_ASM_FMT_NOOPD,            /* SCM_OPCODE_PUSH */
    SCM_ASM_FMT_NOOPD,            /* SCM_OPCODE_MVPUSH */
    SCM_ASM_FMT_NOOPD,            /* SCM_OPCODE_RETURN */
    SCM_ASM_FMT_SI,               /* SCM_OPCODE_PCALL */
    SCM_ASM_FMT_SI,               /* SCM_OPCODE_CALL */
    SCM_ASM_FMT_SI,               /* SCM_OPCODE_TAIL_CALL */
    SCM_ASM_FMT_GV_MOD,           /* SCM_OPCODE_GREF */
    SCM_ASM_FMT_GV_MOD,           /* SCM_OPCODE_GDEF */
    SCM_ASM_FMT_GV_MOD,           /* SCM_OPCODE_GSET */
    SCM_ASM_FMT_SI_SI,            /* SCM_OPCODE_LREF */
    SCM_ASM_FMT_SI_SI,            /* SCM_OPCODE_LSET */
    SCM_ASM_FMT_IOF,              /* SCM_OPCODE_JMP */
    SCM_ASM_FMT_IOF,              /* SCM_OPCODE_JMPT */
    SCM_ASM_FMT_IOF,              /* SCM_OPCODE_JMPF */
    SCM_ASM_FMT_SI_SI,            /* SCM_OPCODE_LBOX */
    SCM_ASM_FMT_SI_SI,            /* SCM_OPCODE_LBREF */
    SCM_ASM_FMT_SI_SI,            /* SCM_OPCODE_LBSET */
    SCM_ASM_FMT_SI_SI_SEQ,        /* SCM_OPCODE_CLOSE */
    SCM_ASM_FMT_SI_SI,            /* SCM_OPCODE_DEMINE */
    SCM_ASM_FMT_SI,               /* SCM_OPCODE_EMINE */
    SCM_ASM_FMT_SI_SI,            /* SCM_OPCODE_EDEMINE */
    SCM_ASM_FMT_SI,               /* SCM_OPCODE_MRVC */
    SCM_ASM_FMT_NOOPD,            /* SCM_OPCODE_MRVE */
    SCM_ASM_FMT_MOD               /* SCM_OPCODE_MODULE */
};

int
scm_asm_inst_format(scm_opcode_t op)
{
  return ((op < SCM_VM_INST_NR_OP) ? asm_inst_fmt_table[op] : (int)op);
}


/***************************************************************************/
/*  AsmModDecls                                                            */
/***************************************************************************/

ScmTypeInfo SCM_ASMMODDECLS_TYPE_INFO = {
  .name                = "asm-mod-decls",
  .flags               = SCM_TYPE_FLG_MMO,
  .obj_print_func      = NULL,
  .obj_size            = sizeof(ScmAsmModDecls),
  .gc_ini_func         = scm_asmmoddecls_gc_initialize,
  .gc_fin_func         = scm_asmmoddecls_gc_finalize,
  .gc_accept_func      = scm_asmmoddecls_gc_accept,
  .gc_accept_func_weak = NULL,
  .extra               = NULL,
};

int
scm_asmmoddecls_initialize(ScmObj asml)
{
  int r;

  scm_assert_obj_type(asml, &SCM_ASMMODDECLS_TYPE_INFO);

  r = eary_init(SCM_ASMMODDECLS_DECLS(asml), sizeof(ScmAsmModDeclElm), 0);
  if (r < 0) return -1;

  return 0;
}

void
scm_asmmoddecls_finalize(ScmObj obj)
{
  scm_assert_obj_type(obj, &SCM_ASMMODDECLS_TYPE_INFO);

  eary_fin(SCM_ASMMODDECLS_DECLS(obj));
}

ScmObj
scm_asmmoddecls_new(scm_mem_type_t mtype)
{
  ScmObj asml = SCM_OBJ_INIT;

  SCM_REFSTK_INIT_REG(&asml);

  asml = scm_alloc_mem(&SCM_ASMMODDECLS_TYPE_INFO, 0, mtype);
  if (scm_obj_null_p(asml)) return SCM_OBJ_NULL;

  if (scm_asmmoddecls_initialize(asml) < 0)
    return SCM_OBJ_NULL;

  return asml;
}

const ScmAsmModDeclElm *
asmmoddecls_elm(ScmObj asml, size_t id)
{
  const ScmAsmModDeclElm *d;

  scm_assert_obj_type(asml, &SCM_ASMMODDECLS_TYPE_INFO);

  if (!scm_asmmoddecls_assigned_p(asml, id)) {
    scm_error("failed to refer module: "
              "specified module declaration id has not assigned", 0);
    return NULL;
  }

  d = eary_idx_to_ptr(SCM_ASMMODDECLS_DECLS(asml), id);
  if (scm_obj_null_p(d->module)) {
    scm_error("failed to refer module: "
              "specified module has not declared", 0);
    return NULL;
  }

  return d;
}

ScmObj
scm_asmmoddecls_module(ScmObj asml, size_t id)
{
  const ScmAsmModDeclElm *d;

  scm_assert_obj_type(asml, &SCM_ASMMODDECLS_TYPE_INFO);

  d = asmmoddecls_elm(asml, id);
  if (d == NULL) return SCM_OBJ_NULL;

  return d->module;
}

int
scm_asmmoddecls_registered_flag(ScmObj asml, size_t id, bool *flg)
{
  const ScmAsmModDeclElm *d;

  scm_assert_obj_type(asml, &SCM_ASMMODDECLS_TYPE_INFO);
  scm_assert(flg);

  d = asmmoddecls_elm(asml, id);
  if (d == NULL) return -1;

  *flg = d->registered;
  return 0;
}

int
scm_asmmoddecls_id(ScmObj asml, ScmObj mod, ssize_t *id)
{
  ScmAsmModDeclElm *d;
  size_t idx;

  scm_assert_obj_type(asml, &SCM_ASMMODDECLS_TYPE_INFO);
  scm_assert(scm_module_p(mod));
  scm_assert(id != NULL);

  EARY_FOR_EACH(SCM_ASMMODDECLS_DECLS(asml), idx, d) {
    if (scm_eq_p(d->module, mod)) {
      *id = (ssize_t)idx;
      return 0;
    }
  }

  *id = (ssize_t)-1;
  return 0;
}

ssize_t
scm_asmmoddecls_assign_id(ScmObj asml)
{
  ScmAsmModDeclElm d = { .module = SCM_OBJ_NULL, .registered = false };
  size_t i;
  int r;

  scm_assert_obj_type(asml, &SCM_ASMMODDECLS_TYPE_INFO);

  i = EARY_SIZE(SCM_ASMMODDECLS_DECLS(asml));
  if (i >= SSIZE_MAX) {
    scm_error("failed to assign module declaration id: "
              "too many module declarations", 0);
    return -1;
  }

  EARY_PUSH(SCM_ASMMODDECLS_DECLS(asml), ScmAsmModDeclElm, d, r);
  if (r < 0) return -1;

  return (ssize_t)i;
}

int
scm_asmmoddecls_register_id(ScmObj asml, size_t id)
{
  const ScmAsmModDeclElm d = { .module = SCM_OBJ_NULL, .registered = false };
  size_t i, s;
  int r;

  scm_assert_obj_type(asml, &SCM_ASMMODDECLS_TYPE_INFO);

  if (id >= SSIZE_MAX) {
    scm_error("failed to register assembler module id: too big", 0);
    return -1;
  }

  s = EARY_SIZE(SCM_ASMMODDECLS_DECLS(asml));
  for (i = id; i > s; i--) {
    EARY_SET(SCM_ASMMODDECLS_DECLS(asml), ScmAsmModDeclElm, i, d, r);
    if (r < 0) return -1;
  }

  if (i == s) {
    EARY_SET(SCM_ASMMODDECLS_DECLS(asml), ScmAsmModDeclElm, s, d, r);
    if (r < 0) return -1;
  }

  return 0;
}

int
scm_asmmoddecls_register_decl(ScmObj asml, size_t id,
                              size_t seq_id, size_t offset, ScmObj mod, bool reg)
{
  ScmAsmModDeclElm *d;

  scm_assert_obj_type(asml, &SCM_ASMMODDECLS_TYPE_INFO);
  scm_assert(scm_module_p(mod));

  if (id >= EARY_SIZE(SCM_ASMMODDECLS_DECLS(asml))) {
    scm_error("failed to register module declaration: "
              "specified module id has not assigned", 0);
    return -1;
  }

  d = eary_idx_to_ptr(SCM_ASMMODDECLS_DECLS(asml), id);
  if (scm_obj_not_null_p(d->module)) {
    scm_error("failed to register module declaration: "
              "specified module id has already declared", 0);
    return -1;
  }

  d->seq_id = seq_id;
  d->offset = offset;
  SCM_WB_SETQ(asml, d->module, mod);
  d->registered = reg;

  return 0;
}

void
scm_asmmoddecls_itr_init(ScmAsmModDeclsItr *itr)
{
  scm_assert(itr != NULL);

  eary_init(&itr->indexes, sizeof(size_t), 0);
  itr->cur = 0;
}

void
scm_asmmoddecls_itr_fin(ScmAsmModDeclsItr *itr)
{
  scm_assert(itr != NULL);

  eary_fin(&itr->indexes);
}

static int
asmmoddecls_itr_push(ScmObj asml, ScmAsmModDeclsItr *itr, size_t idx)
{
  ScmAsmModDeclElm *decl;
  size_t tmp;
  int r;

  scm_assert_obj_type(asml, &SCM_ASMMODDECLS_TYPE_INFO);
  scm_assert(itr != NULL);

  EARY_PUSH(&itr->indexes, size_t, idx, r);
  if (r < 0) return -1;

  decl = eary_idx_to_ptr(SCM_ASMMODDECLS_DECLS(asml), idx);
  for (size_t i = EARY_SIZE(&itr->indexes) - 1; i > 0; i--) {
    size_t *ip = eary_idx_to_ptr(&itr->indexes, i);
    size_t *pip = ip - 1;
    ScmAsmModDeclElm *dp = eary_idx_to_ptr(SCM_ASMMODDECLS_DECLS(asml), *pip);

    if (decl->offset >= dp->offset)
      break;

    tmp = *pip;
    *pip = *ip;
    *ip = tmp;
  }

  return 0;
}

int
scm_asmmoddecls_itr_begin(ScmObj asml, size_t seq_id, ScmAsmModDeclsItr *itr)
{
  ScmAsmModDeclElm *elm;
  size_t idx;
  int r;

  scm_assert_obj_type(asml, &SCM_ASMMODDECLS_TYPE_INFO);
  scm_assert(itr != NULL);

  eary_truncate(&itr->indexes);

  EARY_FOR_EACH(SCM_ASMMODDECLS_DECLS(asml), idx, elm) {
    if (elm->seq_id != seq_id) continue;
    if (scm_obj_null_p(elm->module)) continue;

    r = asmmoddecls_itr_push(asml, itr, idx);
    if (r < 0) return -1;
  }

  itr->cur = 0;
  return 0;
}

bool
scm_asmmoddecls_itr_end_p(ScmObj asml, ScmAsmModDeclsItr *itr)
{
  scm_assert_obj_type(asml, &SCM_ASMMODDECLS_TYPE_INFO);
  scm_assert(itr != NULL);

  return (itr->cur >= EARY_SIZE(&itr->indexes));
}

void
scm_asmmoddecls_itr_next(ScmObj asml, ScmAsmModDeclsItr *itr)
{
  scm_assert_obj_type(asml, &SCM_ASMMODDECLS_TYPE_INFO);
  scm_assert(itr != NULL);

  if (scm_asmmoddecls_itr_end_p(asml, itr))
    return;

  itr->cur++;
  return;
}

size_t
scm_asmmoddecls_itr_id(ScmObj asml, ScmAsmModDeclsItr *itr)
{
  scm_assert_obj_type(asml, &SCM_ASMMODDECLS_TYPE_INFO);
  scm_assert(itr != NULL);
  scm_assert(!scm_asmmoddecls_itr_end_p(asml, itr));

  return *(size_t *)eary_idx_to_ptr(&itr->indexes, itr->cur);
}

static ScmAsmModDeclElm *
asmmoddecls_itr_elm(ScmObj asml, ScmAsmModDeclsItr *itr)
{
  size_t id;

  scm_assert_obj_type(asml, &SCM_ASMMODDECLS_TYPE_INFO);
  scm_assert(itr != NULL);
  scm_assert(!scm_asmmoddecls_itr_end_p(asml, itr));

  EARY_GET(&itr->indexes, size_t, itr->cur, id);
  return eary_idx_to_ptr(SCM_ASMMODDECLS_DECLS(asml), id);
}

size_t
scm_asmmoddecls_itr_offset(ScmObj asml, ScmAsmModDeclsItr *itr)
{
  ScmAsmModDeclElm *d;

  scm_assert_obj_type(asml, &SCM_ASMMODDECLS_TYPE_INFO);
  scm_assert(itr != NULL);
  scm_assert(!scm_asmmoddecls_itr_end_p(asml, itr));

  d = asmmoddecls_itr_elm(asml, itr);
  return d->offset;
}

ScmObj
scm_asmmoddecls_itr_module(ScmObj asml, ScmAsmModDeclsItr *itr)
{
  ScmAsmModDeclElm *d;

  scm_assert_obj_type(asml, &SCM_ASMMODDECLS_TYPE_INFO);
  scm_assert(itr != NULL);
  scm_assert(!scm_asmmoddecls_itr_end_p(asml, itr));

  d = asmmoddecls_itr_elm(asml, itr);
  return d->module;
}

bool
scm_asmmoddecls_itr_registered(ScmObj asml, ScmAsmModDeclsItr *itr)
{
  ScmAsmModDeclElm *d;

  scm_assert_obj_type(asml, &SCM_ASMMODDECLS_TYPE_INFO);
  scm_assert(itr != NULL);
  scm_assert(!scm_asmmoddecls_itr_end_p(asml, itr));

  d = asmmoddecls_itr_elm(asml, itr);
  return d->registered;
}

void
scm_asmmoddecls_gc_initialize(ScmObj obj)
{
  scm_assert_obj_type(obj, &SCM_ASMMODDECLS_TYPE_INFO);

  eary_init(SCM_ASMMODDECLS_DECLS(obj), 0, 0);
}

void
scm_asmmoddecls_gc_finalize(ScmObj obj)
{
  scm_asmmoddecls_finalize(obj);
}

int
scm_asmmoddecls_gc_accept(ScmObj obj, ScmGCRefHandler handler)
{
  ScmAsmModDeclElm *decl;
  size_t idx;
  int rslt;

  scm_assert_obj_type(obj, &SCM_ASMMODDECLS_TYPE_INFO);

  rslt = SCM_GC_REF_HANDLER_VAL_INIT;
  EARY_FOR_EACH(SCM_ASMMODDECLS_DECLS(obj), idx, decl) {
    rslt = SCM_GC_CALL_REF_HANDLER(handler, obj, decl->module);
    if (scm_gc_ref_handler_failure_p(rslt)) return rslt;
  }

  return rslt;
}


/**************************************************************************/
/* Assembler                                                              */
/**************************************************************************/

ScmTypeInfo SCM_ASSEMBLER_TYPE_INFO = {
  .name                = "assembler",
  .flags               = SCM_TYPE_FLG_MMO,
  .obj_print_func      = NULL,
  .obj_size            = sizeof(ScmAssembler),
  .gc_ini_func         = scm_asm_gc_initialize,
  .gc_fin_func         = scm_asm_gc_finalize,
  .gc_accept_func      = scm_asm_gc_accept,
  .gc_accept_func_weak = NULL,
  .extra               = NULL,
};

static ScmObj
asm_get_mod_decls(ScmObj asmb)
{
  scm_assert_obj_type(asmb, &SCM_ASSEMBLER_TYPE_INFO);
  return prog_get_mod_decls(SCM_ASSEMBLER_PROG(asmb));
}

static void
asm_set_mod_decls(ScmObj asmb, ScmObj decls)
{
  scm_assert_obj_type(asmb, &SCM_ASSEMBLER_TYPE_INFO);
  scm_assert(scm_asmmoddecls_p(decls));
  prog_set_mod_decls(SCM_ASSEMBLER_PROG(asmb), decls);
}

static int
scm_asm_initialize_inner(ScmObj asmb, ScmObj prog, size_t id)
{
  ScmObj mod_decls = SCM_OBJ_INIT;
  int r;

  SCM_REFSTK_INIT_REG(&asmb, &prog,
                      &mod_decls);

  scm_assert_obj_type(asmb, &SCM_ASSEMBLER_TYPE_INFO);
  scm_assert(scm_cdprogram_p(prog));
  scm_assert(id <= SSIZE_MAX);


  SCM_SLOT_SETQ(ScmAssembler, asmb, prog, prog);
  SCM_ASSEMBLER(asmb)->id = id;

  SCM_SLOT_SETQ(ScmAssembler, asmb, iseq, scm_cdprogram_iseq(prog, id, false));
  if (scm_obj_null_p(SCM_ASSEMBLER_ISEQ(asmb))) return -1;

  mod_decls = asm_get_mod_decls(asmb);
  if (scm_obj_null_p(mod_decls)) {
    mod_decls = scm_make_asmmoddecls();
    if (scm_obj_null_p(mod_decls))
      return -1;
    asm_set_mod_decls(asmb, mod_decls);
  }

  r = eary_init(SCM_ASSEMBLER_LABEL_DECL(asmb), sizeof(ScmAsmLabelDecl), 0);
  if (r < 0) return -1;

  r = eary_init(SCM_ASSEMBLER_LABEL_REF(asmb), sizeof(ScmAsmLabelRef), 0);
  if (r < 0) return -1;

  return 0;
}

int
scm_asm_initialize(ScmObj asmb, ScmObj prog)
{
  ssize_t id;

  SCM_REFSTK_INIT_REG(&asmb, &prog);

  scm_assert_obj_type(asmb, &SCM_ASSEMBLER_TYPE_INFO);
  scm_assert(scm_obj_null_p(prog) || scm_cdprogram_p(prog));

  if (scm_obj_null_p(prog)) {
    prog = scm_make_cdprogram(SCM_OBJ_NULL, SCM_OBJ_NULL);
    if (scm_obj_null_p(prog)) return -1;

    id = scm_cdprogram_assign_id(prog, 0);
    if (id < 0) return -1;
  }

  return scm_asm_initialize_inner(asmb, prog, 0);
}

void
scm_asm_finalize(ScmObj obj)
{
  scm_assert_obj_type(obj, &SCM_ASSEMBLER_TYPE_INFO);

  SCM_ASSEMBLER(obj)->prog = SCM_OBJ_NULL;
  SCM_ASSEMBLER(obj)->iseq = SCM_OBJ_NULL;
  eary_fin(SCM_ASSEMBLER_LABEL_DECL(obj));
  eary_fin(SCM_ASSEMBLER_LABEL_REF(obj));
}

ScmObj
scm_asm_new(scm_mem_type_t mtype, ScmObj prog)
{
  ScmObj asmb = SCM_OBJ_INIT;

  SCM_REFSTK_INIT_REG(&prog,
                      &asmb);

  scm_assert(scm_obj_null_p(prog) || scm_cdprogram_p(prog));

  asmb = scm_alloc_mem(&SCM_ASSEMBLER_TYPE_INFO, 0, mtype);
  if (scm_obj_null_p(asmb)) return SCM_OBJ_NULL;

  if (scm_asm_initialize(asmb, prog) < 0)
    return SCM_OBJ_NULL;

  return asmb;
}

ScmObj
scm_asm_new_seq(ScmObj asmb, size_t id)
{
  ScmObj new_asm = SCM_OBJ_INIT, iseq = SCM_OBJ_INIT;
  int r;

  SCM_REFSTK_INIT_REG(&asmb,
                      &new_asm);

  scm_assert_obj_type(asmb, &SCM_ASSEMBLER_TYPE_INFO);
  scm_assert(id < SSIZE_MAX);

  if (!scm_cdprogram_assigned_id_p(SCM_ASSEMBLER_PROG(asmb), id)) {
    scm_error("failed to create sequence block: "
              "sepcified sequence id has not assigned", 0);
    return SCM_OBJ_NULL;
  }

  iseq = scm_cdprogram_iseq(SCM_ASSEMBLER_PROG(asmb), id, true);
  if (scm_obj_null_p(iseq)) return SCM_OBJ_NULL;

  new_asm = scm_alloc_mem(&SCM_ASSEMBLER_TYPE_INFO, 0, SCM_MEM_HEAP);
  if (scm_obj_null_p(new_asm)) return SCM_OBJ_NULL;

  r = scm_asm_initialize_inner(new_asm, SCM_ASSEMBLER_PROG(asmb), id);
  if (r < 0) return SCM_OBJ_NULL;

  return new_asm;
}

ScmObj
scm_asm_head_iseq(ScmObj asmb)
{
  scm_assert_obj_type(asmb, &SCM_ASSEMBLER_TYPE_INFO);

  return scm_cdprogram_head_iseq(SCM_ASSEMBLER_PROG(asmb));
}

static inline bool
scm_asm_label_id_assigned_p(ScmObj asmb, size_t id)
{
  scm_assert_obj_type(asmb, &SCM_ASSEMBLER_TYPE_INFO);
  return (id < EARY_SIZE(SCM_ASSEMBLER_LABEL_DECL(asmb)));
}

static int
scm_asm_register_label_ref(ScmObj asmb, size_t offset, size_t label_id)
{
  ScmAsmLabelRef ref = { .offset = offset, .label_id = label_id };
  int r;

  scm_assert_obj_type(asmb, &SCM_ASSEMBLER_TYPE_INFO);

  EARY_PUSH(SCM_ASSEMBLER_LABEL_REF(asmb), ScmAsmLabelRef, ref, r);
  if (r < 0) return -1;

  return 0;
}

static int
scm_asm_calc_iof(const ScmAsmLabelRef *ref, const ScmAsmLabelDecl *decl,
                 int *iof)
{
  ptrdiff_t x;

  scm_assert(ref != NULL);
  scm_assert(decl != NULL);
  scm_assert(iof != NULL);

  if (decl->offset < 0) {
    scm_error("failed to resolve reference to assembler labels: "
              "specified label has not undeclared", 0);
    return -1;
  }

  x = ((ptrdiff_t)decl->offset
       - (ptrdiff_t)(ref->offset + SCM_VM_INST_SZ_OPD_IOF));
  if ((ref->offset > PTRDIFF_MAX - SCM_VM_INST_SZ_OPD_IOF)
      || (x < INT_MIN || INT_MAX < x)) {
    scm_error("failed to resolve reference to assembler labels: "
              "arithmetic overflow", 0);
    return -1;
  }

  *iof = (int)x;
  return 0;
}

static ScmObj
scm_asm_get_module(ScmObj asmb, ScmObj mod)
{
  ScmObj m = SCM_OBJ_NULL;
  int r;

  SCM_REFSTK_INIT_REG(&asmb, &mod,
                      &m);

  scm_assert_obj_type(asmb, &SCM_ASSEMBLER_TYPE_INFO);
  scm_assert(scm_obj_not_null_p(mod));

  if (scm_module_p(mod)) {
    return mod;
  }
  else if (scm_module_name_p(mod)) {
    r = scm_find_registered_module(mod, SCM_CSETTER_L(m));
    if (r < 0) return SCM_OBJ_NULL;

    if (scm_obj_null_p(m)) {
      scm_error("failed to find registered module: no such a module", 1, mod);
      return SCM_OBJ_NULL;
    }

    return m;
  }
  else if (scm_num_integer_p(mod)) {
    size_t i;

    r = scm_integer_to_size_t(mod, &i);
    if (r < 0) return SCM_OBJ_NULL;

    return scm_asmmoddecls_module(asm_get_mod_decls(asmb), i);
  }
  else {
    scm_error("assembler: invalid module specification", 1, mod);
    return SCM_OBJ_NULL;
  }
}

ssize_t
scm_asm_assign_label_id(ScmObj asmb)
{
  ScmAsmLabelDecl d = { .offset = -1 };
  size_t i;
  int r;

  scm_assert_obj_type(asmb, &SCM_ASSEMBLER_TYPE_INFO);

  i = EARY_SIZE(SCM_ASSEMBLER_LABEL_DECL(asmb));
  if (i >= SSIZE_MAX) {
    scm_error("failed to assign assembler label id: too many lables", 0);
    return -1;
  }

  EARY_PUSH(SCM_ASSEMBLER_LABEL_DECL(asmb), ScmAsmLabelDecl, d, r);
  if (r < 0) return -1;

  return (ssize_t)i;
}

int
scm_asm_register_label_id(ScmObj asmb, size_t id)
{
  const ScmAsmLabelDecl d = { .offset = -1 };
  int r;

  scm_assert_obj_type(asmb, &SCM_ASSEMBLER_TYPE_INFO);

  if (id >= SSIZE_MAX) {
    scm_error("failed to register assembler label id: too big", 0);
    return -1;
  }

  for (size_t i = EARY_SIZE(SCM_ASSEMBLER_LABEL_DECL(asmb)); i <= id; i++) {
    EARY_SET(SCM_ASSEMBLER_LABEL_DECL(asmb), ScmAsmLabelDecl, i, d, r);
    if (r < 0) return -1;
  }

  return 0;
}

ssize_t
scm_asm_assign_module_id(ScmObj asmb)
{
  scm_assert_obj_type(asmb, &SCM_ASSEMBLER_TYPE_INFO);
  return scm_asmmoddecls_assign_id(asm_get_mod_decls(asmb));
}

int
scm_asm_register_module_id(ScmObj asmb, size_t id)
{
  scm_assert_obj_type(asmb, &SCM_ASSEMBLER_TYPE_INFO);
  return scm_asmmoddecls_register_id(asm_get_mod_decls(asmb), id);
}

ssize_t
scm_asm_assign_sequence_id(ScmObj asmb)
{
  scm_assert_obj_type(asmb, &SCM_ASSEMBLER_TYPE_INFO);
  return scm_cdprogram_assign_id(SCM_ASSEMBLER_PROG(asmb), -1);
}

int
scm_asm_register_sequence_id(ScmObj asmb, size_t id)
{
  ssize_t i;

  scm_assert_obj_type(asmb, &SCM_ASSEMBLER_TYPE_INFO);
  scm_assert(id <= SSIZE_MAX);

  i = scm_cdprogram_assign_id(SCM_ASSEMBLER_PROG(asmb), (ssize_t)id);
  return ((i < 0) ? -1 : 0);
}

int
scm_asm_push_inst_noopd(ScmObj asmb, scm_opcode_t op)
{
  scm_byte_t inst[SCM_VM_INST_SZ_NOOPD];
  ssize_t r;

  scm_assert_obj_type(asmb, &SCM_ASSEMBLER_TYPE_INFO);

  op = scm_internal_opcode(op);
  scm_vm_inst_set_inst_noopd(inst, op);
  r = scm_iseq_push_inst(SCM_ASSEMBLER_ISEQ(asmb),
                         &inst, sizeof(inst), NULL, 0);
  if (r < 0) return -1;

  return 0;
}

int
scm_asm_push_inst_obj(ScmObj asmb, scm_opcode_t op, ScmObj obj)
{
  static const size_t objs[1] = { SCM_VM_INST_OFFSET_OBJ_1 };
  scm_byte_t inst[SCM_VM_INST_SZ_OBJ];
  ssize_t r;

  scm_assert_obj_type(asmb, &SCM_ASSEMBLER_TYPE_INFO);
  scm_assert(scm_obj_not_null_p(obj));

  op = scm_internal_opcode(op);
  scm_vm_inst_set_inst_obj(inst, op, obj);
  r = scm_iseq_push_inst(SCM_ASSEMBLER_ISEQ(asmb),
                         &inst, sizeof(inst), objs, 1);
  if (r < 0) return -1;

  return 0;
}

int
scm_asm_push_inst_gv_mod(ScmObj asmb,
                         scm_opcode_t op, ScmObj gv, ScmObj mod)
{
  static const size_t objs[2] = {
    SCM_VM_INST_OFFSET_OBJ_OBJ_1, SCM_VM_INST_OFFSET_OBJ_OBJ_2
  };
  scm_byte_t inst[SCM_VM_INST_SZ_OBJ_OBJ];
  ssize_t r;

  SCM_REFSTK_INIT_REG(&asmb, &gv, &mod);

  scm_assert_obj_type(asmb, &SCM_ASSEMBLER_TYPE_INFO);
  scm_assert(scm_obj_not_null_p(gv));
  scm_assert(scm_obj_not_null_p(mod));

  op = scm_internal_opcode(op);
  mod = scm_asm_get_module(asmb, mod);
  if (scm_obj_null_p(mod)) return -1;

  scm_vm_inst_set_inst_obj_obj(inst, op, gv, mod);
  r = scm_iseq_push_inst(SCM_ASSEMBLER_ISEQ(asmb),
                         &inst, sizeof(inst), objs, 2);
  if (r < 0) return -1;

  return 0;
}

int
scm_asm_push_inst_si(ScmObj asmb, scm_opcode_t op, int si)
{
  scm_byte_t inst[SCM_VM_INST_SZ_SI];
  ssize_t r;

  scm_assert_obj_type(asmb, &SCM_ASSEMBLER_TYPE_INFO);

  op = scm_internal_opcode(op);
  scm_vm_inst_set_inst_si(inst, op, si);
  r = scm_iseq_push_inst(SCM_ASSEMBLER_ISEQ(asmb),
                         &inst, sizeof(inst), NULL, 0);
  if (r < 0) return -1;

  return 0;
}

int
scm_asm_push_inst_si_si(ScmObj asmb, scm_opcode_t op, int si1, int si2)
{
  scm_byte_t inst[SCM_VM_INST_SZ_SI_SI];
  ssize_t r;

  scm_assert_obj_type(asmb, &SCM_ASSEMBLER_TYPE_INFO);

  op = scm_internal_opcode(op);
  scm_vm_inst_set_inst_si_si(inst, op, si1, si2);
  r = scm_iseq_push_inst(SCM_ASSEMBLER_ISEQ(asmb),
                         &inst, sizeof(inst), NULL, 0);
  if (r < 0) return -1;

  return 0;
}

int
scm_asm_push_inst_si_si_seq(ScmObj asmb,
                            scm_opcode_t op, int si1, int si2, ScmObj seq)
{
  static const size_t objs[1] = { SCM_VM_INST_OFFSET_SI_SI_OBJ_3 };
  scm_byte_t inst[SCM_VM_INST_SZ_SI_SI_OBJ];
  ssize_t r;

  SCM_REFSTK_INIT_REG(&asmb, &seq);

  scm_assert_obj_type(asmb, &SCM_ASSEMBLER_TYPE_INFO);

  if (!scm_assembler_p(seq)) {
    scm_error("failed to assemble instruction: invalid operand", 1, seq);
    return -1;
  }
  else if (!scm_eq_p(SCM_ASSEMBLER_PROG(asmb), SCM_ASSEMBLER_PROG(seq))) {
    scm_error("failed to assemble instruction: invalid operand", 1, seq);
    return -1;
  }

  seq = scm_asm_iseq(seq);

  op = scm_internal_opcode(op);
  scm_vm_inst_set_inst_si_si_obj(inst, op, si1, si2, seq);
  r = scm_iseq_push_inst(SCM_ASSEMBLER_ISEQ(asmb),
                         &inst, sizeof(inst), objs, 1);
  if (r < 0) return -1;

  return 0;
}

int
scm_asm_push_inst_iof(ScmObj asmb, scm_opcode_t op, bool label, ...)
{
  va_list arg;
  int ret;

  scm_assert_obj_type(asmb, &SCM_ASSEMBLER_TYPE_INFO);

  va_start(arg, label);
  if (label) {
    ret = scm_asm_push_inst_rlid(asmb, op, va_arg(arg, size_t));
  }
  else {
    scm_byte_t inst[SCM_VM_INST_SZ_IOF];
    op = scm_internal_opcode(op);
    scm_vm_inst_set_inst_iof(inst, op, va_arg(arg, int));
    ssize_t r = scm_iseq_push_inst(SCM_ASSEMBLER_ISEQ(asmb),
                                   &inst, sizeof(inst), NULL, 0);
    if (r < 0) ret = -1;
    else       ret = 0;
  }
  va_end(arg);

  return ret;
}

int
scm_asm_push_inst_mod(ScmObj asmb, scm_opcode_t op, ScmObj mod)
{
  scm_assert_obj_type(asmb, &SCM_ASSEMBLER_TYPE_INFO);
  scm_assert(scm_obj_not_null_p(mod));

  mod = scm_asm_get_module(asmb, mod);
  if (scm_obj_null_p(mod)) return -1;

  return scm_asm_push_inst_obj(asmb, op, mod);
}

int
scm_asm_push_inst_rlid(ScmObj asmb, scm_opcode_t op, size_t id)
{
  const ScmAsmLabelDecl *decl;
  size_t cur;
  int r, iof;

  scm_assert_obj_type(asmb, &SCM_ASSEMBLER_TYPE_INFO);

  if (!scm_asm_label_id_assigned_p(asmb, id)) {
    scm_error("failed to assemble pseudo instruction: "
              "specified label has not undeclared", 0);
    return -1;
  }

  cur = scm_iseq_length(SCM_ASSEMBLER_ISEQ(asmb)) + SCM_VM_INST_OFFSET_OPD;

  decl = eary_idx_to_ptr(SCM_ASSEMBLER_LABEL_DECL(asmb), id);
  if (decl->offset < 0) {
    r = scm_asm_register_label_ref(asmb, cur, id);
    if (r < 0) return -1;
    iof = 0;
  }
  else {
    ScmAsmLabelRef ref = { .offset = cur, .label_id = id };
    r = scm_asm_calc_iof(&ref, decl, &iof);
    if (r < 0) return -1;
  }

  r = scm_asm_push_inst_iof(asmb, op, false, (int)iof);
  if (r < 0) return -1;

  return 0;
}

int
scm_asm_push_pinst_label(ScmObj asmb, scm_opcode_t op, size_t id)
{
  ScmAsmLabelDecl *decl;
  size_t cur;

  scm_assert_obj_type(asmb, &SCM_ASSEMBLER_TYPE_INFO);

  if (id >= EARY_SIZE(SCM_ASSEMBLER_LABEL_DECL(asmb))) {
    scm_error("failed to assemble pseudo instruction: "
              "specified label has not assigned", 0);
    return -1;
  }

  decl = eary_idx_to_ptr(SCM_ASSEMBLER_LABEL_DECL(asmb), id);
  if (decl->offset >= 0) {
    scm_error("failed to assemble pseudo instruction: "
              "specified lable has already declared", 0);
    return -1;
  }

  cur = scm_iseq_length(SCM_ASSEMBLER_ISEQ(asmb));
  if (cur > SSIZE_MAX) {
    scm_error("failed to assemble pseudo instruction: "
              "instructin sequence too big", 0);
    return -1;
  }

  decl->offset = (ssize_t)cur;

  return 0;
}

ScmObj
asm_declmod_to_module(ScmObj mod, bool reg)
{
  scm_assert(scm_module_specifier_p(mod));

  if (scm_module_p(mod)) {
    return mod;
  }
  else if (reg){
    ScmObj m = SCM_OBJ_INIT;

    int r = scm_find_registered_module(mod, SCM_CSETTER_L(m));
    if (r < 0) return SCM_OBJ_NULL;

    if (scm_obj_null_p(m)) {
      scm_error("failed to assemble pseudo instruction: "
                "specified module has not registered", 0);
      return SCM_OBJ_NULL;
    }

    return m;
  }
  else {
    return scm_make_module(mod, false);
  }
}

int
scm_asm_push_pinst_declmod(ScmObj asmb, scm_opcode_t op,
                           size_t id, ScmObj mod, bool reg)
{
  size_t seq_id, offset;
  int r;

  SCM_REFSTK_INIT_REG(&asmb, &mod);

  scm_assert_obj_type(asmb, &SCM_ASSEMBLER_TYPE_INFO);
  scm_assert(scm_module_specifier_p(mod));

  mod = asm_declmod_to_module(mod, reg);
  if (scm_obj_null_p(mod)) return -1;

  seq_id = SCM_ASSEMBLER_ID(asmb);
  offset = scm_iseq_length(SCM_ASSEMBLER_ISEQ(asmb));
  r = scm_asmmoddecls_register_decl(asm_get_mod_decls(asmb), id,
                                    seq_id, offset, mod, reg);
  if (r < 0) return -1;

  return 0;
}

int
scm_asm_push_pinst_sequence(ScmObj asmb, scm_opcode_t op, size_t id)
{
  int r;

  SCM_REFSTK_INIT_REG(&asmb);

  scm_assert_obj_type(asmb, &SCM_ASSEMBLER_TYPE_INFO);

  if (!scm_cdprogram_assigned_id_p(SCM_ASSEMBLER_PROG(asmb), id)) {
    scm_error("failed to assemble pseudo instruction: "
              "specified sequence id has not assigned", 0);
    return -1;
  }

  r = scm_asm_commit(asmb);
  if (r < 0) return -1;

  SCM_ASSEMBLER(asmb)->id = id;
  SCM_ASSEMBLER(asmb)->iseq
    = scm_cdprogram_iseq(SCM_ASSEMBLER_PROG(asmb), id, true);
  if (scm_obj_null_p(SCM_ASSEMBLER_ISEQ(asmb)))
    return -1;

  return 0;
}

static inline int
chk_operand_obj(ScmObj obj)
{
  if (scm_obj_null_p(obj)) {
    scm_error("failed to push a instruction to ISeq: invalid operand",
              1, obj);
    return -1;
  }

  return 0;
}

int
scm_asm_push_va(ScmObj asmb, scm_opcode_t op, va_list operands)
{
  ScmObj opd_obj1 = SCM_OBJ_INIT, opd_obj2 = SCM_OBJ_INIT;
  size_t opd_label_id, opd_module_id, opd_sequence_id;
  int r, opd_si1, opd_si2;
  bool label, reg;

  SCM_REFSTK_INIT_REG(&asmb,
                      &opd_obj1, &opd_obj2);

  scm_assert(scm_assembler_p(asmb));
  /* scm_assert(op >= 0); */

  switch (scm_asm_inst_format(op)) {
  case SCM_ASM_FMT_NOOPD:
    return scm_asm_push_inst_noopd(asmb, op);
    break;
  case SCM_ASM_FMT_OBJ:
    opd_obj1 = va_arg(operands, ScmObj);
    r = chk_operand_obj(opd_obj1);
    if (r < 0) return -1;
    return scm_asm_push_inst_obj(asmb, op, opd_obj1);
    break;
  case SCM_ASM_FMT_GV_MOD:
    opd_obj1 = va_arg(operands, ScmObj);
    opd_obj2 = va_arg(operands, ScmObj);
    r = chk_operand_obj(opd_obj1);
    if (r < 0) return -1;
    r = chk_operand_obj(opd_obj2);
    if (r < 0) return -1;
    return scm_asm_push_inst_gv_mod(asmb, op, opd_obj1, opd_obj2);
    break;
  case SCM_ASM_FMT_SI:
    opd_si1 = va_arg(operands, int);
    return scm_asm_push_inst_si(asmb, op, opd_si1);
    break;
  case SCM_ASM_FMT_SI_SI:
    opd_si1 = va_arg(operands, int);
    opd_si2 = va_arg(operands, int);
    return scm_asm_push_inst_si_si(asmb, op, opd_si1, opd_si2);
    break;
  case SCM_ASM_FMT_SI_SI_SEQ:
    opd_si1 = va_arg(operands, int);
    opd_si2 = va_arg(operands, int);
    opd_obj1 = va_arg(operands, ScmObj);
    r = chk_operand_obj(opd_obj1);
    if (r < 0) return -1;
    return scm_asm_push_inst_si_si_seq(asmb, op, opd_si1, opd_si2, opd_obj1);
    break;
  case SCM_ASM_FMT_IOF:
    label = (bool)va_arg(operands, int);
    if (label) {
      opd_label_id = va_arg(operands, size_t);
      return scm_asm_push_inst_iof(asmb, op, label, opd_label_id);
    }
    else {
      opd_si1 = va_arg(operands, int);
      return scm_asm_push_inst_iof(asmb, op, label, opd_si1);
    }
    break;
  case SCM_ASM_FMT_MOD:
    opd_obj1 = va_arg(operands, ScmObj);
    r = chk_operand_obj(opd_obj1);
    if (r < 0) return -1;
    return scm_asm_push_inst_mod(asmb, op, opd_obj1);
    break;
  case SCM_ASM_PI_LABEL:
    opd_label_id = va_arg(operands, size_t);
    return scm_asm_push_pinst_label(asmb, op, opd_label_id);
    break;
  case SCM_ASM_PI_DECLMOD:
    opd_module_id = va_arg(operands, size_t);
    opd_obj1 = va_arg(operands, ScmObj);
    reg = (bool)va_arg(operands, int);
    return scm_asm_push_pinst_declmod(asmb, op, opd_module_id, opd_obj1, reg);
    break;
  case SCM_ASM_PI_SEQUENCE:
    opd_sequence_id = va_arg(operands, size_t);
    return scm_asm_push_pinst_sequence(asmb, op, opd_sequence_id);
    break;
  default:
    scm_error("failed to assemble: unknown instruction format", 0);
    return -1;
    break;
  }
}

int
scm_asm_push(ScmObj asmb, scm_opcode_t op, ...)
{
  va_list operands;
  int r;

  scm_assert(scm_assembler_p(asmb));
  /* scm_assert(op >= 0); */

  va_start(operands, op);
  r = scm_asm_push_va(asmb, op, operands);
  va_end(operands);

  return r;
}

int
scm_asm_resolve_label_ref(ScmObj asmb)
{
  const ScmAsmLabelDecl *decl;
  const ScmAsmLabelRef *ref;
  size_t idx;
  int r, iof;

  scm_assert_obj_type(asmb, &SCM_ASSEMBLER_TYPE_INFO);

  EARY_FOR_EACH(SCM_ASSEMBLER_LABEL_REF(asmb), idx, ref) {
    decl = eary_idx_to_ptr(SCM_ASSEMBLER_LABEL_DECL(asmb), ref->label_id);
    r = scm_asm_calc_iof(ref, decl, &iof);
    if (r < 0) return -1;

    scm_asm_update_iof(SCM_ASSEMBLER_ISEQ(asmb),
                       scm_iseq_to_ip(SCM_ASSEMBLER_ISEQ(asmb)) + ref->offset,
                       iof);
  }

  eary_truncate(SCM_ASSEMBLER_LABEL_REF(asmb));
  return 0;
}

void
scm_asm_clear_labels(ScmObj asmb)
{
  scm_assert_obj_type(asmb, &SCM_ASSEMBLER_TYPE_INFO);
  if (EARY_SIZE(SCM_ASSEMBLER_LABEL_REF(asmb)) == 0)
    eary_truncate(SCM_ASSEMBLER_LABEL_DECL(asmb));
}

int
scm_asm_commit(ScmObj asmb)
{
  int r;

  scm_assert(scm_assembler_p(asmb));

  r = scm_asm_resolve_label_ref(asmb);
  if (r < 0) return -1;

  scm_asm_clear_labels(asmb);
  return 0;
}

void
scm_asm_gc_initialize(ScmObj obj)
{
  scm_assert_obj_type(obj, &SCM_ASSEMBLER_TYPE_INFO);

  SCM_ASSEMBLER(obj)->prog = SCM_OBJ_NULL;
  SCM_ASSEMBLER(obj)->iseq = SCM_OBJ_NULL;
  eary_init(SCM_ASSEMBLER_LABEL_DECL(obj), 0, 0);
  eary_init(SCM_ASSEMBLER_LABEL_REF(obj), 0, 0);
}

void
scm_asm_gc_finalize(ScmObj obj)
{
  scm_asm_finalize(obj);
}

int
scm_asm_gc_accept(ScmObj obj, ScmGCRefHandler handler)
{
  int rslt = SCM_GC_REF_HANDLER_VAL_INIT;

  scm_assert_obj_type(obj, &SCM_ASSEMBLER_TYPE_INFO);

  rslt = SCM_GC_CALL_REF_HANDLER(handler, obj, SCM_ASSEMBLER_PROG(obj));
  if (scm_gc_ref_handler_failure_p(rslt)) return rslt;

  rslt = SCM_GC_CALL_REF_HANDLER(handler, obj, SCM_ASSEMBLER_ISEQ(obj));
  if (scm_gc_ref_handler_failure_p(rslt)) return rslt;

  return rslt;
}


/**************************************************************************/
/* Disassembler                                                           */
/**************************************************************************/

ScmTypeInfo SCM_DISASSEMBLER_TYPE_INFO = {
  .name                = "disassembler",
  .flags               = SCM_TYPE_FLG_MMO,
  .obj_print_func      = NULL,
  .obj_size            = sizeof(ScmDisassembler),
  .gc_ini_func         = scm_disasm_gc_initialize,
  .gc_fin_func         = scm_disasm_gc_finalize,
  .gc_accept_func      = scm_disasm_gc_accept,
  .gc_accept_func_weak = NULL,
  .extra               = NULL,
};

static ScmObj
disasm_get_mod_decls(ScmObj disasm)
{
  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);
  return prog_get_mod_decls(SCM_DISASSEMBLER_PROG(disasm));
}

static void
scm_disasm_token_clear(ScmDisasmToken *tk)
{
  scm_assert(tk != NULL);

  tk->type = SCM_DISASM_TK_INST;
  tk->fmt = SCM_ASM_FMT_NOOPD;
  tk->op = SCM_OPCODE_NOP;
}

static int
scm_disasm_push_label_decl(ScmObj disasm, size_t offset)
{
  ScmAsmLabelDecl d = { .offset = (ssize_t)offset };
  ScmAsmLabelDecl *ary;
  size_t idx;
  int r;

  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);
  scm_assert(offset <= SSIZE_MAX);

  EARY_PUSH(SCM_DISASSEMBLER_LABEL_DECLS(disasm), ScmAsmLabelDecl, d, r);
  if (r < 0) return -1;

  idx = EARY_SIZE(SCM_DISASSEMBLER_LABEL_DECLS(disasm)) - 1;
  ary = EARY_HEAD(SCM_DISASSEMBLER_LABEL_DECLS(disasm));
  while (idx > 0 && d.offset < ary[idx - 1].offset)
    idx--;

  if (idx > 0 && d.offset == ary[idx - 1].offset) {
    EARY_POP(SCM_DISASSEMBLER_LABEL_DECLS(disasm), ScmAsmLabelDecl, d);
    return 0;
  }

  for (size_t i = EARY_SIZE(SCM_DISASSEMBLER_LABEL_DECLS(disasm)); i > idx; i--)
    ary[i] = ary[i - 1];
  ary[idx] = d;

  return 0;
}

static int
scm_disasm_acc_labels_iof(ScmObj disasm, scm_byte_t *ip)
{
  ssize_t offset;
  int r, iof;

  SCM_REFSTK_INIT_REG(&disasm);

  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);
  scm_assert(ip != NULL);

  scm_vm_inst_fetch_opd_iof(&ip, &iof);
  offset = scm_iseq_ip_to_offset(SCM_DISASSEMBLER_ISEQ(disasm), ip + iof);
  if (offset < 0) return -1;

  r = scm_disasm_push_label_decl(disasm, (size_t)offset);
  if (r < 0) return -1;

  return 0;
}

static int
scm_disasm_preproc(ScmObj disasm)
{
  scm_byte_t *ip;
  int r;

  SCM_REFSTK_INIT_REG(&disasm);

  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);

  if (scm_obj_null_p(SCM_DISASSEMBLER_ISEQ(disasm)))
    return 0;

  if (scm_obj_not_null_p(SCM_DISASSEMBLER_MOD_DECLS(disasm))) {
    r = scm_asmmoddecls_itr_begin(SCM_DISASSEMBLER_MOD_DECLS(disasm),
                                  SCM_DISASSEMBLER_SEQ_ID(disasm),
                                  SCM_DISASSEMBLER_DECL_MOD_ITR(disasm));
    if (r < 0) return -1;
  }

  eary_truncate(SCM_DISASSEMBLER_LABEL_DECLS(disasm));

  ip = scm_iseq_to_ip(SCM_DISASSEMBLER_ISEQ(disasm));
  while (scm_iseq_ip_in_range_p(SCM_DISASSEMBLER_ISEQ(disasm), ip)) {
    scm_opcode_t op = scm_external_opcode(scm_vm_inst_fetch_op(&ip));
    switch (scm_asm_inst_format(op)) {
    case SCM_ASM_FMT_IOF:
      r = scm_disasm_acc_labels_iof(disasm, ip);
      if (r < 0) return -1;
      break;
    default:
      /* do nothing */
      break;
    }
    ip += scm_vm_inst_op2opdsize(op);
  }

  return 0;
}

static int
label_decl_cmp(const void *x, const void *y)
{
  const ScmAsmLabelDecl *a = x, *b = y;

  if (a->offset < b->offset)
    return -1;
  else if (a->offset == b->offset)
    return 0;
  else
    return 1;
}

static int
scm_disasm_setup_token_inst_gv_mod(ScmObj disasm)
{
  int r;

  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);

  SCM_WB_EXP(disasm,
             memcpy(&SCM_DISASSEMBLER_TOKEN(disasm)->opd.obj_obj.o,
                    SCM_DISASSEMBLER_IP(disasm),
                    SCM_VM_INST_SZ_OPD_OBJ_OBJ));

  SCM_DISASSEMBLER_ADD_IP(disasm, SCM_VM_INST_SZ_OPD_OBJ_OBJ);

  if (scm_gloc_p(SCM_DISASSEMBLER_TOKEN(disasm)->opd.obj_obj.o.o1)) {
    SCM_WB_SETQ(disasm,
                SCM_DISASSEMBLER_TOKEN(disasm)->opd.obj_obj.o.o1,
                scm_gloc_symbol(SCM_DISASSEMBLER_TOKEN(disasm)->opd.obj_obj.o.o1));
  }

  if (scm_obj_null_p(SCM_DISASSEMBLER_MOD_DECLS(disasm))) {
    SCM_DISASSEMBLER_TOKEN(disasm)->opd.obj_obj.mod_id = -1;
  }
  else {
    r = scm_asmmoddecls_id(SCM_DISASSEMBLER_MOD_DECLS(disasm),
                           SCM_DISASSEMBLER_TOKEN(disasm)->opd.obj_obj.o.o2,
                           &SCM_DISASSEMBLER_TOKEN(disasm)->opd.obj_obj.mod_id);
    if (r < 0) return -1;
  }

  return 0;
}

static int
scm_disasm_setup_token_inst_iof(ScmObj disasm)
{
  ScmAsmLabelDecl key, *p;
  int iof;

  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);

  memcpy(&SCM_DISASSEMBLER_TOKEN(disasm)->opd.iof.o,
         SCM_DISASSEMBLER_IP(disasm),
         SCM_VM_INST_SZ_OPD_IOF);

  SCM_DISASSEMBLER_ADD_IP(disasm, SCM_VM_INST_SZ_OPD_IOF);

  iof = SCM_DISASSEMBLER_TOKEN(disasm)->opd.iof.o.o1;
  key.offset = scm_iseq_ip_to_offset(SCM_DISASSEMBLER_ISEQ(disasm),
                                     SCM_DISASSEMBLER_IP(disasm) + iof);
  if (key.offset < 0) return -1;
  p = eary_bsearch(SCM_DISASSEMBLER_LABEL_DECLS(disasm),
                   &key, label_decl_cmp);
  scm_assert(p != NULL);      /* must not happen */
  SCM_DISASSEMBLER_TOKEN(disasm)->opd.iof.label_id =
    (size_t)(p - (ScmAsmLabelDecl *)EARY_HEAD(SCM_DISASSEMBLER_LABEL_DECLS(disasm)));

  return 0;
}

static int
scm_disasm_setup_token_inst_mod(ScmObj disasm)
{
  int r;

  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);

  SCM_WB_EXP(disasm,
             memcpy(&SCM_DISASSEMBLER_TOKEN(disasm)->opd.obj.o,
                    SCM_DISASSEMBLER_IP(disasm),
                    SCM_VM_INST_SZ_OPD_OBJ));

  SCM_DISASSEMBLER_ADD_IP(disasm, SCM_VM_INST_SZ_OPD_OBJ);

  if (scm_obj_null_p(SCM_DISASSEMBLER_MOD_DECLS(disasm))) {
    SCM_DISASSEMBLER_TOKEN(disasm)->opd.obj.mod_id = -1;
  }
  else {
    r = scm_asmmoddecls_id(SCM_DISASSEMBLER_MOD_DECLS(disasm),
                           SCM_DISASSEMBLER_TOKEN(disasm)->opd.obj.o.o1,
                           &SCM_DISASSEMBLER_TOKEN(disasm)->opd.obj.mod_id);
    if (r < 0) return -1;
  }

  return 0;
}

static int
scm_disasm_setup_token_inst_si_si_seq(ScmObj disasm)
{
  ssize_t id;

  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);

  SCM_WB_EXP(disasm,
             memcpy(&SCM_DISASSEMBLER_TOKEN(disasm)->opd.si_si_obj.o,
                    SCM_DISASSEMBLER_IP(disasm),
                    SCM_VM_INST_SZ_SI_SI_OBJ));

  SCM_DISASSEMBLER_ADD_IP(disasm, SCM_VM_INST_SZ_OPD_SI_SI_OBJ);

  id = scm_cdprogram_id(SCM_DISASSEMBLER_PROG(disasm),
                        SCM_DISASSEMBLER_TOKEN(disasm)->opd.si_si_obj.o.o3);
  if (id < 0) {
    scm_error("failed to dissassemble: unknown sequence", 0);
    return -1;
  }

  SCM_DISASSEMBLER_TOKEN(disasm)->opd.si_si_obj.seq_id = (size_t)id;

  return 0;
}

static bool
scm_disasm_end_of_sequence_p(ScmObj disasm)
{
  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);
  scm_assert(scm_obj_not_null_p(SCM_DISASSEMBLER_ISEQ(disasm)));

  if (scm_iseq_ip_in_range_p(SCM_DISASSEMBLER_ISEQ(disasm),
                             SCM_DISASSEMBLER_IP(disasm)))
    return false;

  if (scm_obj_not_null_p(SCM_DISASSEMBLER_MOD_DECLS(disasm))) {
    if (!scm_asmmoddecls_itr_end_p(SCM_DISASSEMBLER_MOD_DECLS(disasm),
                                   SCM_DISASSEMBLER_DECL_MOD_ITR(disasm)))
      return false;
  }

  if (SCM_DISASSEMBLER_DECL_LABEL_IDX(disasm)
      < EARY_SIZE(SCM_DISASSEMBLER_LABEL_DECLS(disasm)))
    return false;

  return true;
}

static void
scm_disasm_next_iseq(ScmObj disasm)
{
  ScmObj iseq = SCM_OBJ_INIT;
  size_t id;

  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);

  id = SCM_DISASSEMBLER_SEQ_ID(disasm);
  iseq = scm_cdprogram_next_iseq(SCM_DISASSEMBLER_PROG(disasm), &id);
  SCM_DISASSEMBLER_SET_SEQ_ID(disasm, id);
  SCM_DISASSEMBLER_SET_ISEQ(disasm, iseq);
  SCM_DISASSEMBLER_SET_IP(disasm, NULL);
}

static int
scm_disasm_setup_token_inst(ScmObj disasm)
{
  scm_opcode_t op;
  int r;

  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);
  scm_assert(scm_iseq_ip_in_range_p(SCM_DISASSEMBLER_ISEQ(disasm),
                                    SCM_DISASSEMBLER_IP(disasm)));

  SCM_DISASSEMBLER_TOKEN(disasm)->type = SCM_DISASM_TK_INST;

  op = scm_vm_inst_fetch_op(&SCM_DISASSEMBLER_IP(disasm));
  op = scm_external_opcode(op);
  SCM_DISASSEMBLER_TOKEN(disasm)->fmt = scm_asm_inst_format(op);
  SCM_DISASSEMBLER_TOKEN(disasm)->op = op;

  switch (SCM_DISASSEMBLER_TOKEN(disasm)->fmt) {
  case SCM_ASM_FMT_NOOPD:
    memcpy(&SCM_DISASSEMBLER_TOKEN(disasm)->opd.noopd.o,
           SCM_DISASSEMBLER_IP(disasm),
           SCM_VM_INST_SZ_OPD_NOOPD);
    SCM_DISASSEMBLER_ADD_IP(disasm, SCM_VM_INST_SZ_OPD_NOOPD);
    break;
  case SCM_ASM_FMT_OBJ:
    SCM_WB_EXP(disasm,
               memcpy(&SCM_DISASSEMBLER_TOKEN(disasm)->opd.obj.o,
                      SCM_DISASSEMBLER_IP(disasm),
                      SCM_VM_INST_SZ_OPD_OBJ));
    SCM_DISASSEMBLER_ADD_IP(disasm, SCM_VM_INST_SZ_OPD_OBJ);
    break;
  case SCM_ASM_FMT_GV_MOD:
    r = scm_disasm_setup_token_inst_gv_mod(disasm);
    if (r < 0) return -1;
    break;
  case SCM_ASM_FMT_SI:
    memcpy(&SCM_DISASSEMBLER_TOKEN(disasm)->opd.si.o,
           SCM_DISASSEMBLER_IP(disasm),
           SCM_VM_INST_SZ_OPD_SI);
    SCM_DISASSEMBLER_ADD_IP(disasm, SCM_VM_INST_SZ_OPD_SI);
    break;
  case SCM_ASM_FMT_SI_SI:
    memcpy(&SCM_DISASSEMBLER_TOKEN(disasm)->opd.si_si.o,
           SCM_DISASSEMBLER_IP(disasm),
           SCM_VM_INST_SZ_OPD_SI_SI);
    SCM_DISASSEMBLER_ADD_IP(disasm, SCM_VM_INST_SZ_OPD_SI_SI);
    break;
  case SCM_ASM_FMT_SI_SI_SEQ:
    r = scm_disasm_setup_token_inst_si_si_seq(disasm);
    break;
  case SCM_ASM_FMT_IOF:
    r = scm_disasm_setup_token_inst_iof(disasm);
    if (r < 0) return -1;
    break;
  case SCM_ASM_FMT_MOD:
    r = scm_disasm_setup_token_inst_mod(disasm);
    if (r < 0) return -1;
    break;
  default:
    scm_error("failed to disassemble: unknown instruction format", 0);
    return -1;
    break;
  }

  if (scm_disasm_end_of_sequence_p(disasm))
    scm_disasm_next_iseq(disasm);

  return 0;
}

static int
scm_disasm_setup_token_pinst_label(ScmObj disasm)
{
  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);

  SCM_DISASSEMBLER_TOKEN(disasm)->type = SCM_DISASM_TK_PSEUDO;
  SCM_DISASSEMBLER_TOKEN(disasm)->fmt = scm_asm_inst_format(SCM_ASM_PI_LABEL);
  SCM_DISASSEMBLER_TOKEN(disasm)->op = SCM_ASM_PI_LABEL;
  SCM_DISASSEMBLER_TOKEN(disasm)->opd.label.id = SCM_DISASSEMBLER_DECL_LABEL_IDX(disasm);
  SCM_DISASSEMBLER_INC_DECL_LABEL_IDX(disasm);

  if (scm_disasm_end_of_sequence_p(disasm))
    scm_disasm_next_iseq(disasm);

  return 0;
}

static int
scm_disasm_setup_token_pinst_declmod(ScmObj disasm, size_t id)
{
  int r;

  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);

  SCM_DISASSEMBLER_TOKEN(disasm)->type = SCM_DISASM_TK_PSEUDO;
  SCM_DISASSEMBLER_TOKEN(disasm)->fmt = scm_asm_inst_format(SCM_ASM_PI_DECLMOD);
  SCM_DISASSEMBLER_TOKEN(disasm)->op = SCM_ASM_PI_DECLMOD;
  SCM_DISASSEMBLER_TOKEN(disasm)->opd.declmod.id = id;

  SCM_WB_SETQ(disasm, SCM_DISASSEMBLER_TOKEN(disasm)->opd.declmod.mod,
              scm_asmmoddecls_module(SCM_DISASSEMBLER_MOD_DECLS(disasm), id));
  if (scm_obj_null_p(SCM_DISASSEMBLER_TOKEN(disasm)->opd.declmod.mod)) return -1;

  r = scm_asmmoddecls_registered_flag(SCM_DISASSEMBLER_MOD_DECLS(disasm),
                                      id,
                                      &SCM_DISASSEMBLER_TOKEN(disasm)->opd.declmod.reg);
  if (r < 0) return -1;

  if (scm_disasm_end_of_sequence_p(disasm))
    scm_disasm_next_iseq(disasm);

  return 0;
}

static int
scm_disasm_setup_token_pinst_sequence(ScmObj disasm)
{
  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);

  SCM_DISASSEMBLER_TOKEN(disasm)->type = SCM_DISASM_TK_PSEUDO;
  SCM_DISASSEMBLER_TOKEN(disasm)->fmt = scm_asm_inst_format(SCM_ASM_PI_SEQUENCE);
  SCM_DISASSEMBLER_TOKEN(disasm)->op = SCM_ASM_PI_SEQUENCE;
  SCM_DISASSEMBLER_TOKEN(disasm)->opd.sequence.id = SCM_DISASSEMBLER_SEQ_ID(disasm);

  if (scm_disasm_end_of_sequence_p(disasm))
    scm_disasm_next_iseq(disasm);

  return 0;
}

static int
scm_disasm_setup_token_inner(ScmObj disasm)
{
  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);
  scm_assert(SCM_DISASSEMBLER_IP(disasm) != NULL);

  if (SCM_DISASSEMBLER_MOD_DECLS(disasm)
      && !scm_asmmoddecls_itr_end_p(SCM_DISASSEMBLER_MOD_DECLS(disasm),
                                    SCM_DISASSEMBLER_DECL_MOD_ITR(disasm))) {
    size_t offset = scm_asmmoddecls_itr_offset(SCM_DISASSEMBLER_MOD_DECLS(disasm),
                                               SCM_DISASSEMBLER_DECL_MOD_ITR(disasm));
    if (offset == (size_t)scm_iseq_ip_to_offset(SCM_DISASSEMBLER_ISEQ(disasm),
                                                SCM_DISASSEMBLER_IP(disasm))) {
      size_t id = scm_asmmoddecls_itr_id(SCM_DISASSEMBLER_MOD_DECLS(disasm),
                                         SCM_DISASSEMBLER_DECL_MOD_ITR(disasm));
      scm_asmmoddecls_itr_next(SCM_DISASSEMBLER_MOD_DECLS(disasm),
                               SCM_DISASSEMBLER_DECL_MOD_ITR(disasm));
      return scm_disasm_setup_token_pinst_declmod(disasm, id);
    }
  }

  if (SCM_DISASSEMBLER_DECL_LABEL_IDX(disasm)
      < EARY_SIZE(SCM_DISASSEMBLER_LABEL_DECLS(disasm))) {
    ScmAsmLabelDecl *p = eary_idx_to_ptr(SCM_DISASSEMBLER_LABEL_DECLS(disasm),
                                         SCM_DISASSEMBLER_DECL_LABEL_IDX(disasm));
    if (p->offset == scm_iseq_ip_to_offset(SCM_DISASSEMBLER_ISEQ(disasm),
                                           SCM_DISASSEMBLER_IP(disasm)))
      return scm_disasm_setup_token_pinst_label(disasm);
  }

  return scm_disasm_setup_token_inst(disasm);
}

static inline int
scm_disasm_setup_token(ScmObj disasm)
{
  SCM_REFSTK_INIT_REG(&disasm);

  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);

  if (SCM_DISASSEMBLER_TOKEN(disasm)->type == SCM_DISASM_TK_END)
    return 0;

  if (SCM_DISASSEMBLER_IP(disasm) == NULL) {
    if (scm_obj_null_p(SCM_DISASSEMBLER_ISEQ(disasm))) {
      SCM_DISASSEMBLER_TOKEN(disasm)->type = SCM_DISASM_TK_END;
      return 0;
    }
    else {
      SCM_DISASSEMBLER_SET_IP(disasm,
                              scm_iseq_to_ip(SCM_DISASSEMBLER_ISEQ(disasm)));
      scm_disasm_preproc(disasm);
      SCM_DISASSEMBLER_SET_DECL_LABEL_IDX(disasm, 0);
      scm_disasm_setup_token_pinst_sequence(disasm);
      return 0;
    }
  }

  return scm_disasm_setup_token_inner(disasm);
}

int
scm_disasm_initialize(ScmObj disasm, ScmObj prog)
{
  int r;

  SCM_REFSTK_INIT_REG(&disasm, &prog);

  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);
  scm_assert(scm_cdprogram_p(prog));

  SCM_SLOT_SETQ(ScmDisassembler, disasm, prog, prog);
  SCM_SLOT_SETQ(ScmDisassembler, disasm, iseq, SCM_OBJ_NULL);
  SCM_SLOT_SETQ(ScmDisassembler, disasm, mod_decls, SCM_OBJ_NULL);

  SCM_DISASSEMBLER_SET_SEQ_ID(disasm, 0);
  SCM_DISASSEMBLER_SET_IP(disasm, NULL);
  SCM_DISASSEMBLER_SET_DECL_LABEL_IDX(disasm, 0);

  scm_asmmoddecls_itr_init(SCM_DISASSEMBLER_DECL_MOD_ITR(disasm));

  r = eary_init(SCM_DISASSEMBLER_LABEL_DECLS(disasm),
                sizeof(ScmAsmLabelDecl), 0);
  if (r < 0) return -1;

  scm_disasm_rewind(disasm);

  return 0;
}

void
scm_disasm_finalize(ScmObj disasm)
{
  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);

  eary_fin(SCM_DISASSEMBLER_LABEL_DECLS(disasm));

  scm_asmmoddecls_itr_fin(SCM_DISASSEMBLER_DECL_MOD_ITR(disasm));
}

ScmObj
scm_disasm_new(scm_mem_type_t mtype, ScmObj prog)
{
  ScmObj disasm;

  SCM_REFSTK_INIT_REG(&prog,
                      &disasm);

  scm_assert(scm_cdprogram_p(prog));

  disasm = scm_alloc_mem(&SCM_DISASSEMBLER_TYPE_INFO, 0, mtype);
  if (scm_obj_null_p(disasm)) return SCM_OBJ_NULL;

  if (scm_disasm_initialize(disasm, prog) < 0)
    return SCM_OBJ_NULL;

  return disasm;
}

int
scm_disasm_next(ScmObj disasm)
{
  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);

  return scm_disasm_setup_token(disasm);
}

void
scm_disasm_rewind(ScmObj disasm)
{
  SCM_REFSTK_INIT_REG(&disasm);

  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);

  SCM_SLOT_SETQ(ScmDisassembler, disasm, iseq,
                scm_cdprogram_head_iseq(SCM_DISASSEMBLER_PROG(disasm)));
  SCM_SLOT_SETQ(ScmDisassembler, disasm, mod_decls,
                disasm_get_mod_decls(disasm));
  SCM_DISASSEMBLER_SET_SEQ_ID(disasm, 0);

  SCM_DISASSEMBLER_SET_DECL_LABEL_IDX(disasm, 0);

  scm_disasm_token_clear(SCM_DISASSEMBLER_TOKEN(disasm));

  if (scm_obj_not_null_p(SCM_DISASSEMBLER_ISEQ(disasm))) {
    SCM_DISASSEMBLER_SET_IP(disasm,
                            scm_iseq_to_ip(SCM_DISASSEMBLER_ISEQ(disasm)));
    scm_disasm_preproc(disasm);
    scm_disasm_setup_token(disasm);
  }
}

void
scm_disasm_gc_initialize(ScmObj obj)
{
  scm_assert_obj_type(obj, &SCM_DISASSEMBLER_TYPE_INFO);

  SCM_DISASSEMBLER(obj)->prog = SCM_OBJ_NULL;
  SCM_DISASSEMBLER(obj)->iseq = SCM_OBJ_NULL;
  SCM_DISASSEMBLER(obj)->mod_decls = SCM_OBJ_NULL;
  eary_init(SCM_DISASSEMBLER_LABEL_DECLS(obj), 0, 0);
  SCM_DISASSEMBLER_TOKEN(obj)->type = SCM_DISASM_TK_END;
}

void
scm_disasm_gc_finalize(ScmObj obj)
{
  scm_disasm_finalize(obj);
}

int
scm_disasm_gc_accept(ScmObj obj, ScmGCRefHandler handler)
{
  int rslt = SCM_GC_REF_HANDLER_VAL_INIT;

  scm_assert_obj_type(obj, &SCM_DISASSEMBLER_TYPE_INFO);

  rslt = SCM_GC_CALL_REF_HANDLER(handler, obj, SCM_DISASSEMBLER_PROG(obj));
  if (scm_gc_ref_handler_failure_p(rslt)) return rslt;

  rslt = SCM_GC_CALL_REF_HANDLER(handler, obj, SCM_DISASSEMBLER_ISEQ(obj));
  if (scm_gc_ref_handler_failure_p(rslt)) return rslt;

  rslt = SCM_GC_CALL_REF_HANDLER(handler, obj, SCM_DISASSEMBLER_MOD_DECLS(obj));
  if (scm_gc_ref_handler_failure_p(rslt)) return rslt;

  if (SCM_DISASSEMBLER_TOKEN(obj)->type != SCM_DISASM_TK_END) {
    switch (SCM_DISASSEMBLER_TOKEN(obj)->fmt) {
    case SCM_ASM_FMT_OBJ:
      rslt = SCM_GC_CALL_REF_HANDLER(handler,
                                     obj,
                                     SCM_DISASSEMBLER_TOKEN(obj)->opd.obj.o.o1);
      if (scm_gc_ref_handler_failure_p(rslt)) return rslt;
      break;
    case SCM_ASM_FMT_GV_MOD:
      rslt = SCM_GC_CALL_REF_HANDLER(handler,
                                     obj,
                                     SCM_DISASSEMBLER_TOKEN(obj)->opd.obj_obj.o.o1);
      if (scm_gc_ref_handler_failure_p(rslt)) return rslt;
      rslt = SCM_GC_CALL_REF_HANDLER(handler,
                                     obj,
                                     SCM_DISASSEMBLER_TOKEN(obj)->opd.obj_obj.o.o2);
      if (scm_gc_ref_handler_failure_p(rslt)) return rslt;
      break;
    case SCM_ASM_FMT_SI_SI_SEQ:
      rslt = SCM_GC_CALL_REF_HANDLER(handler,
                                     obj,
                                     SCM_DISASSEMBLER_TOKEN(obj)->opd.si_si_obj.o.o3);
      if (scm_gc_ref_handler_failure_p(rslt)) return rslt;
      break;
    case SCM_ASM_FMT_MOD:
      rslt = SCM_GC_CALL_REF_HANDLER(handler,
                                     obj,
                                     SCM_DISASSEMBLER_TOKEN(obj)->opd.obj.o.o1);
      if (scm_gc_ref_handler_failure_p(rslt)) return rslt;
      break;
    case SCM_ASM_PI_DECLMOD:
      rslt = SCM_GC_CALL_REF_HANDLER(handler,
                                     obj,
                                     SCM_DISASSEMBLER_TOKEN(obj)->opd.declmod.mod);
      if (scm_gc_ref_handler_failure_p(rslt)) return rslt;
    default:
      break;
    }
  }

  return rslt;
}


/**************************************************************************/
/* Assemble/Disassemble                                                   */
/**************************************************************************/

static struct {
  scm_opcode_t code;
  const char *mne;
} opcode2mnemonic_tbl[] = {
  { SCM_OPCODE_NOP,         "nop" },
  { SCM_OPCODE_HALT,        "halt" },
  { SCM_OPCODE_INT,         "int" },
  { SCM_OPCODE_CFRAME,      "cframe" },
  { SCM_OPCODE_EFRAME,      "eframe" },
  { SCM_OPCODE_EPOP,        "epop" },
  { SCM_OPCODE_ESHIFT,      "eshift" },
  { SCM_OPCODE_IMMVAL,      "immval" },
  { SCM_OPCODE_PUSH,        "push" },
  { SCM_OPCODE_MVPUSH,      "mvpush" },
  { SCM_OPCODE_RETURN,      "return" },
  { SCM_OPCODE_PCALL,       "pcall" },
  { SCM_OPCODE_CALL,        "call" },
  { SCM_OPCODE_TAIL_CALL,   "tcall" },
  { SCM_OPCODE_GREF,        "gref" },
  { SCM_OPCODE_GDEF,        "gdef" },
  { SCM_OPCODE_GSET,        "gset" },
  { SCM_OPCODE_LREF,        "lref" },
  { SCM_OPCODE_LSET,        "lset" },
  { SCM_OPCODE_JMP,         "jmp" },
  { SCM_OPCODE_JMPT,        "jmpt" },
  { SCM_OPCODE_JMPF,        "jmpf" },
  { SCM_OPCODE_LBOX,        "lbox" },
  { SCM_OPCODE_LBREF,       "lbref" },
  { SCM_OPCODE_LBSET,       "lbset" },
  { SCM_OPCODE_CLOSE,       "close" },
  { SCM_OPCODE_DEMINE,      "demine" },
  { SCM_OPCODE_EMINE,       "emine" },
  { SCM_OPCODE_EDEMINE,     "edemine" },
  { SCM_OPCODE_MRVC,        "mrvc" },
  { SCM_OPCODE_MRVE,        "mrve" },
  { SCM_OPCODE_MODULE,      "module" },
  { SCM_ASM_PI_LABEL,       "label" },
  { SCM_ASM_PI_DECLMOD,     "declmod" },
  { SCM_ASM_PI_SEQUENCE,    "sequence" },
};

static int
scm_asm_sym2opcode(ScmObj op)
{
  SCM_REFSTK_INIT_REG(&op);

  scm_assert(!scm_obj_null_p(op));

  if (scm_num_integer_p(op)) {
    scm_sword_t cd;
    int r = scm_integer_to_sword(op, &cd);
    if (r < 0) return -1;

    if ((0 <= cd && cd < SCM_VM_INST_NR_OP)
        || (SCM_ASM_PI_CODE_START <= cd && cd < SCM_ASM_PI_CODE_END)) {
      return (int)cd;
    }
    else if (cd == SCM_ASM_PI_LABEL) {
      return (int)cd;
    }
    else {
      scm_error("assembler: invalid opcode", 1, op);
      return -1;
    }
  }
  else if (scm_symbol_p(op)) {
    ssize_t rslt;
    char mne[32];
    char *p = scm_symbol_to_cstr(op, mne, sizeof(mne));
    if (p == NULL) return -1;

    rslt = scm_asm_mnemonic2opcode(mne);
    if (rslt < 0) {
      scm_error("assembler: unknown mnemonic", 1, op);
      return -1;
    }

    return (int)rslt;
  }
  else {
    scm_error("assembler: invalid opcode", 1, op);
    return -1;
  }
}

static int
scm_asm_chk_nr_operand(const ScmObj *inst, size_t n, size_t min, size_t max)
{
  ScmObj lst = SCM_OBJ_INIT;

  scm_assert(n == 0 || inst != NULL);

  if (n == 0 || (n - 1) < min) {
    lst = scm_list_cv(inst, n);
    if (scm_obj_not_null_p(lst))
      scm_error("assembler: too few operands", 1, lst);
    return -1;
  }
  else if ((n - 1) > max) {
    lst = scm_list_cv(inst, n);
    if (scm_obj_not_null_p(lst))
      scm_error("assembler: too many operands", 1, lst);
    return -1;
  }

  return 0;
}

static int
scm_asm_cnv_operand_to_si(ScmObj operand, ScmObj operator, int *si)
{
  scm_sword_t val;
  int r;

  scm_assert(scm_obj_not_null_p(operand));
  scm_assert(scm_obj_not_null_p(operator));
  scm_assert(si != NULL);

  if (!scm_num_integer_p(operand)) {
    scm_error("assembler: invalid operands: integer requried",
              2, operator, operand);
    return -1;
  }

  r = scm_integer_to_sword(operand, &val);
  if (r < 0) return -1;

  if (val < INT_MIN || INT_MAX < val) {
    scm_error("assembler: invalid operands: out of range",
              2, operator, operand);
    return -1;
  }

  *si = (int)val;
  return 0;
}

static int
scm_asm_asm_inst_noopd(ScmObj asmb,
                       scm_opcode_t opcode, const ScmObj *inst, size_t n)
{
  int rslt;

  SCM_REFSTK_INIT_REG(&asmb);

  scm_assert(scm_assembler_p(asmb));
  scm_assert(n == 0 || inst != NULL);

  rslt = scm_asm_chk_nr_operand(inst, n, 0, 0);
  if (rslt < 0) return -1;

  rslt = scm_asm_push_inst_noopd(asmb, opcode);
  if (rslt < 0) return -1;

  return 0;
}

static int
scm_asm_asm_inst_obj(ScmObj asmb,
                     scm_opcode_t opcode, const ScmObj *inst, size_t n)
{
  int rslt;

  SCM_REFSTK_INIT_REG(&asmb);

  scm_assert(scm_assembler_p(asmb));
  scm_assert(n == 0 || inst != NULL);

  rslt = scm_asm_chk_nr_operand(inst, n, 1, 1);
  if (rslt < 0) return -1;

  rslt = scm_asm_push_inst_obj(asmb, opcode, inst[1]);
  if (rslt < 0) return -1;

  return 0;
}

static int
scm_asm_asm_inst_gv_mod(ScmObj asmb,
                        scm_opcode_t opcode, const ScmObj *inst, size_t n)
{
  ScmObj arg1 = SCM_OBJ_INIT, arg2 = SCM_OBJ_INIT;
  int rslt;

  SCM_REFSTK_INIT_REG(&asmb,
                      &arg1, &arg2);

  scm_assert(scm_assembler_p(asmb));
  scm_assert(n == 0 || inst != NULL);

  rslt = scm_asm_chk_nr_operand(inst, n, 2, 2);
  if (rslt < 0) return -1;

  rslt = scm_asm_push_inst_gv_mod(asmb, opcode, inst[1], inst[2]);
  if (rslt < 0) return -1;

  return 0;
}

static int
scm_asm_asm_inst_si(ScmObj asmb,
                    scm_opcode_t opcode, const ScmObj *inst, size_t n)
{
  int rslt, val;

  SCM_REFSTK_INIT_REG(&asmb);

  scm_assert(scm_assembler_p(asmb));
  scm_assert(n == 0 || inst != NULL);

  rslt = scm_asm_chk_nr_operand(inst, n, 1, 1);
  if (rslt < 0) return -1;

  rslt = scm_asm_cnv_operand_to_si(inst[1], inst[0], &val);
  if (rslt < 0) return -1;

  rslt = scm_asm_push_inst_si(asmb, opcode, val);
  if (rslt < 0) return -1;

  return 0;
}

static int
scm_asm_asm_inst_si_si(ScmObj asmb,
                       scm_opcode_t opcode, const ScmObj *inst, size_t n)
{
  int rslt, val1, val2;

  SCM_REFSTK_INIT_REG(&asmb);

  scm_assert(scm_assembler_p(asmb));
  scm_assert(n == 0 || inst != NULL);

  rslt = scm_asm_chk_nr_operand(inst, n, 2, 2);
  if (rslt < 0) return -1;

  rslt = scm_asm_cnv_operand_to_si(inst[1], inst[0], &val1);
  if (rslt < 0) return -1;

  rslt = scm_asm_cnv_operand_to_si(inst[2], inst[0], &val2);
  if (rslt < 0) return -1;

  rslt = scm_asm_push_inst_si_si(asmb, opcode, val1, val2);
  if (rslt < 0) return -1;

  return 0;
}

static ScmObj
asm_asm_extract_seq_id_from_seq_opd(ScmObj asmb, ScmObj opc, ScmObj opd)
{
  if (scm_num_integer_p(opd)) {
    return opd;
  }
  else if (scm_pair_p(opd) && scm_symbol_p(scm_car(opd))) {
    ScmObj x = scm_cdr(opd);
    if (scm_pair_p(x) && scm_num_integer_p(scm_car(x)))
      return scm_car(x);
  }

  scm_error("assembler: invalid operand", 2, opc, opd);
  return SCM_OBJ_NULL;
}

static ScmObj
asm_asm_handle_seq_opd(ScmObj asmb, scm_opcode_t opcode, ScmObj opc, ScmObj opd)
{
  ScmObj new_asm = SCM_OBJ_INIT;
  size_t id;
  int rslt;

  SCM_REFSTK_INIT_REG(&asmb, &opc, &opd,
                      &new_asm);

  scm_assert(scm_assembler_p(asmb));

  if (scm_assembler_p(opd))
    return opd;

  opd = asm_asm_extract_seq_id_from_seq_opd(asmb, opc, opd);
  if (scm_obj_null_p(opd)) return SCM_OBJ_NULL;

  if (scm_num_negative_p(opd)) {
    scm_error("assembler: sequence id: positive integer required",
              2, opc, opd);
    return SCM_OBJ_NULL;
  }

  rslt = scm_integer_to_size_t(opd, &id);
  if (rslt < 0) return SCM_OBJ_NULL;

  rslt = scm_asm_register_sequence_id(asmb, id);
  if (rslt < 0) return SCM_OBJ_NULL;

  return scm_asm_new_seq(asmb, id);
}

static int
scm_asm_asm_inst_si_si_seq(ScmObj asmb,
                           scm_opcode_t opcode, const ScmObj *inst, size_t n)
{
  ScmObj opd3 = SCM_OBJ_INIT;
  int rslt, val1, val2;

  SCM_REFSTK_INIT_REG(&asmb,
                      &opd3);

  scm_assert(scm_assembler_p(asmb));
  scm_assert(n == 0 || inst != NULL);

  rslt = scm_asm_chk_nr_operand(inst, n, 3, 3);
  if (rslt < 0) return -1;

  rslt = scm_asm_cnv_operand_to_si(inst[1], inst[0], &val1);
  if (rslt < 0) return -1;

  rslt = scm_asm_cnv_operand_to_si(inst[2], inst[0], &val2);
  if (rslt < 0) return -1;

  opd3 = asm_asm_handle_seq_opd(asmb, opcode, inst[0], inst[3]);
  if (scm_obj_null_p(opd3)) return -1;

  rslt = scm_asm_push_inst_si_si_seq(asmb, opcode, val1, val2, opd3);
  if (rslt < 0) return -1;

  return 0;
}

static int
scm_asm_asm_inst_iof_raw(ScmObj asmb,
                         scm_opcode_t opcode, const ScmObj *inst, size_t n)
{
  int rslt;

  SCM_REFSTK_INIT_REG(&asmb);

  scm_assert(scm_assembler_p(asmb));
  scm_assert(n == 0 || inst != NULL);

  rslt = scm_asm_chk_nr_operand(inst, n, 2, 2);
  if (rslt < 0) return -1;

  if (scm_true_p(inst[1])) {
    size_t id;

    if (!scm_num_integer_p(inst[2]) || !scm_num_positive_p(inst[2])) {
      scm_error("assembler: label id: positive integer required",
                2, inst[0], inst[1]);
      return -1;
    }

    rslt = scm_integer_to_size_t(inst[2], &id);
    if (rslt < 0) return -1;

    rslt = scm_asm_register_label_id(asmb, id);
    if (rslt < 0) return -1;

    rslt = scm_asm_push_inst_iof(asmb, opcode, true, id);
    if (rslt < 0) return -1;
  }
  else {
    int iof;

    rslt = scm_asm_cnv_operand_to_si(inst[2], inst[0], &iof);
    if (rslt < 0) return -1;

    rslt = scm_asm_push_inst_iof(asmb, opcode, false, iof);
    if (rslt < 0) return -1;
  }

  return 0;
}

static int
scm_asm_asm_inst_iof(ScmObj asmb,
                     scm_opcode_t opcode, const ScmObj *inst, size_t n)
{
  ScmObj new_inst[3] = { SCM_OBJ_INIT, SCM_OBJ_INIT, SCM_OBJ_INIT };
  int rslt;

  SCM_REFSTK_INIT_REG(&asmb);
  SCM_REFSTK_REG_ARY(new_inst, 3);

  scm_assert(scm_assembler_p(asmb));
  scm_assert(n == 0 || inst != NULL);

  rslt = scm_asm_chk_nr_operand(inst, n, 1, 2);
  if (rslt < 0) return -1;

  if (scm_boolean_p(inst[1]))
    return scm_asm_asm_inst_iof_raw(asmb, opcode, inst, n);

  new_inst[0] = inst[0];
  if (scm_pair_p(inst[1])) {
    new_inst[1] = SCM_TRUE_OBJ;
    new_inst[2] = scm_list_ref(inst[1], 1);
    return scm_asm_asm_inst_iof_raw(asmb, opcode, new_inst, 3);
  }
  else if (scm_num_integer_p(inst[1])) {
    new_inst[1] = SCM_FALSE_OBJ;
    new_inst[2] = inst[1];
    return scm_asm_asm_inst_iof_raw(asmb, opcode, new_inst, 3);
  }
  else {
    scm_error("assembler: invalid operands", 2, inst[0], inst[1]);
    return -1;
  }

  return 0;
}

static int
scm_asm_asm_inst_mod(ScmObj asmb,
                     scm_opcode_t opcode, const ScmObj *inst, size_t n)
{
  int rslt;

  SCM_REFSTK_INIT_REG(&asmb);

  scm_assert(scm_assembler_p(asmb));
  scm_assert(n == 0 || inst != NULL);

  rslt = scm_asm_chk_nr_operand(inst, n, 1, 1);
  if (rslt < 0) return -1;

  rslt = scm_asm_push_inst_mod(asmb, opcode, inst[1]);
  if (rslt < 0) return -1;

  return 0;
}

static int
scm_asm_asm_pinst_label(ScmObj asmb,
                        scm_opcode_t opcode, const ScmObj *inst, size_t n)

{
  size_t id;
  int rslt;

  SCM_REFSTK_INIT_REG(&asmb);

  scm_assert(scm_assembler_p(asmb));
  scm_assert(opcode >= SCM_ASM_PI_CODE_START);
  scm_assert(n == 0 || inst != NULL);

  rslt = scm_asm_chk_nr_operand(inst, n, 1, 1);
  if (rslt < 0) return -1;

  if (!scm_num_integer_p(inst[1]) || !scm_num_positive_p(inst[1])) {
    scm_error("assembler: label id: positive integer requried",
              2, inst[0], inst[1]);
    return -1;
  }

  rslt = scm_integer_to_size_t(inst[1], &id);
  if (rslt < 0) return -1;

  rslt = scm_asm_register_label_id(asmb, id);
  if (rslt < 0) return -1;

  rslt = scm_asm_push_pinst_label(asmb, opcode, id);
  if (rslt < 0) return -1;

  return 0;
}

static int
scm_asm_asm_pinst_declmod(ScmObj asmb,
                          scm_opcode_t opcode, const ScmObj *inst, size_t n)

{
  size_t id;
  int rslt;

  SCM_REFSTK_INIT_REG(&asmb);

  scm_assert(scm_assembler_p(asmb));
  scm_assert(opcode >= SCM_ASM_PI_CODE_START);
  scm_assert(n == 0 || inst != NULL);

  rslt = scm_asm_chk_nr_operand(inst, n, 3, 3);
  if (rslt < 0) return -1;

  if (!scm_num_integer_p(inst[1]) || !scm_num_positive_p(inst[1])) {
    scm_error("assembler: module id: positive integer requried",
              2, inst[0], inst[1]);
    return -1;
  }

  if (!scm_boolean_p(inst[3])) {
    scm_error("assembler: registered flag: boolean requried",
              2, inst[0], inst[3]);
    return -1;
  }

  rslt = scm_integer_to_size_t(inst[1], &id);
  if (rslt < 0) return -1;

  rslt = scm_asm_register_module_id(asmb, id);
  if (rslt < 0) return -1;

  rslt = scm_asm_push_pinst_declmod(asmb, opcode,
                                    id, inst[2], scm_true_p(inst[3]));
  if (rslt < 0) return -1;

  return 0;
}

int
scm_asm_asm_pinst_sequence(ScmObj asmb,
                           scm_opcode_t opcode, const ScmObj *inst, size_t n)
{
  size_t id;
  int rslt;

  SCM_REFSTK_INIT_REG(&asmb);

  scm_assert(scm_assembler_p(asmb));
  scm_assert(opcode >= SCM_ASM_PI_CODE_START);
  scm_assert(n == 0 || inst != NULL);

  rslt = scm_asm_chk_nr_operand(inst, n, 1, 1);
  if (rslt < 0) return -1;

  if (!scm_num_integer_p(inst[1]) || !scm_num_positive_p(inst[1])) {
    scm_error("assembler: sequence id: positive integer requried",
              2, inst[0], inst[1]);
    return -1;
  }

  rslt = scm_integer_to_size_t(inst[1], &id);
  if (rslt < 0) return -1;

  rslt = scm_asm_register_sequence_id(asmb, id);
  if (rslt < 0) return -1;

  rslt = scm_asm_push_pinst_sequence(asmb, opcode, id);
  if (rslt < 0) return -1;

  return 0;
}

int
scm_asm_assemble_1inst_cv(ScmObj asmb, const ScmObj *inst, size_t n)
{
  int opcode;

  SCM_REFSTK_INIT_REG(&asmb);

  scm_assert(scm_assembler_p(asmb));
  scm_assert(n == 0 || inst != NULL);

  opcode = scm_asm_sym2opcode(inst[0]);
  if (opcode < 0) return -1;

  switch (scm_asm_inst_format((scm_opcode_t)opcode)) {
  case SCM_ASM_FMT_NOOPD:
    return scm_asm_asm_inst_noopd(asmb, (scm_opcode_t)opcode, inst, n);
    break;
  case SCM_ASM_FMT_OBJ:
    return scm_asm_asm_inst_obj(asmb, (scm_opcode_t)opcode, inst, n);
    break;
  case SCM_ASM_FMT_GV_MOD:
    return scm_asm_asm_inst_gv_mod(asmb, (scm_opcode_t)opcode, inst, n);
    break;
  case SCM_ASM_FMT_SI:
    return scm_asm_asm_inst_si(asmb, (scm_opcode_t)opcode, inst, n);
    break;
  case SCM_ASM_FMT_SI_SI:
    return scm_asm_asm_inst_si_si(asmb, (scm_opcode_t)opcode, inst, n);
    break;
  case SCM_ASM_FMT_SI_SI_SEQ:
    return scm_asm_asm_inst_si_si_seq(asmb, (scm_opcode_t)opcode, inst, n);
    break;
  case SCM_ASM_FMT_IOF:
    return scm_asm_asm_inst_iof(asmb, (scm_opcode_t)opcode, inst, n);
    break;
  case SCM_ASM_FMT_MOD:
    return scm_asm_asm_inst_mod(asmb, (scm_opcode_t)opcode, inst, n);
    break;
  case SCM_ASM_PI_LABEL:
    return scm_asm_asm_pinst_label(asmb, (scm_opcode_t)opcode, inst, n);
    break;
  case SCM_ASM_PI_DECLMOD:
    return scm_asm_asm_pinst_declmod(asmb, (scm_opcode_t)opcode, inst, n);
    break;
  case SCM_ASM_PI_SEQUENCE:
    return scm_asm_asm_pinst_sequence(asmb, (scm_opcode_t)opcode, inst, n);
    break;
  default:
    scm_assert(false);
    break;
  }

  return -1;
}

static ssize_t
scm_asm_asm_inst_list_to_cv(ScmObj lst, ScmObj *cv, size_t n)
{
  ScmObj x = SCM_OBJ_INIT;
  size_t cnt;

  scm_assert(scm_nil_p(lst) || scm_pair_p(lst));
  scm_assert(cv != NULL);

  for (x = lst, cnt = 0; scm_pair_p(x) && cnt < n; x = scm_cdr(x), cnt++)
    cv[cnt] = scm_car(x);

  if (scm_pair_p(x)) {
    scm_error("assembler: too many operands", 1, lst);
    return -1;
  }
  else if (cnt == 0) {
    scm_error("assembler: invalid assembler format", 1, lst);
    return -1;
  }

  scm_assert(cnt <= SSIZE_MAX);
  return (ssize_t)cnt;
}

int
scm_asm_assemble_1inst(ScmObj asmb, ScmObj inst)
{
  ScmObj cv[SCM_ASM_NR_OPD_MAX + 1];  /* + 1 はオペレータ分 */
  ssize_t n;

  SCM_REFSTK_INIT_REG(&asmb, &inst);

  for (int i = 0; i < SCM_ASM_NR_OPD_MAX + 1; i++) cv[i] = SCM_OBJ_NULL;
  SCM_REFSTK_REG_ARY(cv, SCM_ASM_NR_OPD_MAX + 1);

  scm_assert(scm_assembler_p(asmb));
  scm_assert(scm_pair_p(inst) || scm_nil_p(inst));

  n = scm_asm_asm_inst_list_to_cv(inst, cv, SCM_ASM_NR_OPD_MAX + 1);
  if (n < 0) return -1;

  return scm_asm_assemble_1inst_cv(asmb, cv, (size_t)n);
}

int
scm_asm_assemble(ScmObj asmb, ScmObj lst)
{
  ScmObj cur = SCM_OBJ_INIT;

  SCM_REFSTK_INIT_REG(&asmb, &lst,
                      &cur);

  scm_assert(scm_assembler_p(asmb));
  scm_assert(scm_pair_p(lst) || scm_nil_p(lst));

  for (cur = lst; scm_pair_p(cur); cur = scm_cdr(cur)) {
    int r = scm_asm_assemble_1inst(asmb, scm_car(cur));
    if (r < 0) return -1;
  }

  return 0;
}

static ScmObj
scm_disasm_dasm_inst_noopd(ScmObj disasm, ScmObj mne)
{
  scm_assert(scm_disassembler_p(disasm));
  scm_assert(scm_symbol_p(mne));

  return scm_cons(mne, SCM_NIL_OBJ);
}

static ScmObj
scm_disasm_dasm_inst_obj(ScmObj disasm, ScmObj mne)
{
  scm_assert(scm_disassembler_p(disasm));
  scm_assert(scm_symbol_p(mne));

  return scm_list(2, mne, scm_disasm_opd_obj__1(disasm));
}

static ScmObj
scm_disasm_dasm_inst_gv_mod(ScmObj disasm, ScmObj mne)
{
  ScmObj mod = SCM_OBJ_INIT;

  SCM_REFSTK_INIT_REG(&mne,
                      &mod);

  scm_assert(scm_disassembler_p(disasm));
  scm_assert(scm_symbol_p(mne));

  if (scm_disasm_opd_gv_mod__mod_id(disasm) >= 0) {
    mod = scm_make_number_from_size_t((size_t)scm_disasm_opd_gv_mod__mod_id(disasm));
    if (scm_obj_null_p(mod)) return SCM_OBJ_NULL;
  }
  else {
    mod = scm_disasm_opd_gv_mod__2(disasm);
  }

  return scm_list(3, mne, scm_disasm_opd_gv_mod__1(disasm), mod);
}

static ScmObj
scm_disasm_dasm_inst_si(ScmObj disasm, ScmObj mne)
{
  ScmObj num1 = SCM_OBJ_INIT;

  SCM_REFSTK_INIT_REG(&mne,
                      &num1);

  scm_assert(scm_disassembler_p(disasm));
  scm_assert(scm_symbol_p(mne));

  num1 = scm_make_number_from_sword(scm_disasm_opd_si__1(disasm));
  if (scm_obj_null_p(num1)) return SCM_OBJ_NULL;

  return scm_list(2, mne, num1);
}

static ScmObj
scm_disasm_dasm_inst_si_si(ScmObj disasm, ScmObj mne)
{
  ScmObj num1 = SCM_OBJ_INIT, num2 = SCM_OBJ_INIT;

  SCM_REFSTK_INIT_REG(&mne,
                      &num1, &num2);

  scm_assert(scm_disassembler_p(disasm));
  scm_assert(scm_symbol_p(mne));

  num1 = scm_make_number_from_sword(scm_disasm_opd_si_si__1(disasm));
  if (scm_obj_null_p(num1)) return SCM_OBJ_NULL;

  num2 = scm_make_number_from_sword(scm_disasm_opd_si_si__2(disasm));
  if (scm_obj_null_p(num2)) return SCM_OBJ_NULL;

  return scm_list(3, mne, num1, num2);
}

static ScmObj
scm_disasm_dasm_inst_si_si_seq(ScmObj disasm, ScmObj mne)
{
  ScmObj num1 = SCM_OBJ_INIT, num2 = SCM_OBJ_INIT, seq = SCM_OBJ_INIT;
  ScmObj sym = SCM_OBJ_NULL, id = SCM_OBJ_INIT;

  SCM_REFSTK_INIT_REG(&mne,
                      &num1, &num2, &seq,
                      &sym, &id);

  scm_assert(scm_disassembler_p(disasm));
  scm_assert(scm_symbol_p(mne));

  num1 = scm_make_number_from_sword(scm_disasm_opd_si_si_seq__1(disasm));
  if (scm_obj_null_p(num1)) return SCM_OBJ_NULL;

  num2 = scm_make_number_from_sword(scm_disasm_opd_si_si_seq__1(disasm));
  if (scm_obj_null_p(num2)) return SCM_OBJ_NULL;

  sym = scm_make_symbol_from_cstr("sequence", SCM_ENC_SRC);
  if (scm_obj_null_p(sym)) return SCM_OBJ_NULL;

  id = scm_make_number_from_size_t(scm_disasm_opd_si_si_seq__seq_id(disasm));
  if (scm_obj_null_p(id)) return SCM_OBJ_NULL;

  seq = scm_list(2, sym, id);
  if (scm_obj_null_p(seq)) return SCM_OBJ_NULL;

  return scm_list(4, mne, num1, num2, seq);
}

static ScmObj
scm_disasm_dasm_inst_iof(ScmObj disasm, ScmObj mne)
{
  ScmObj sym = SCM_OBJ_INIT, inst = SCM_OBJ_INIT;

  SCM_REFSTK_INIT_REG(&mne,
                      &sym, &inst);

  scm_assert(scm_disassembler_p(disasm));
  scm_assert(scm_symbol_p(mne));

  sym = scm_make_symbol_from_cstr("label", SCM_ENC_SRC);
  if (scm_obj_null_p(sym)) return SCM_OBJ_NULL;

  inst = scm_make_number_from_size_t(scm_disasm_opd_iof__label_id(disasm));
  if (scm_obj_null_p(inst)) return SCM_OBJ_NULL;

  inst = scm_list(2, sym, inst);
  if (scm_obj_null_p(inst)) return SCM_OBJ_NULL;

  return scm_list(2, mne, inst);
}

static ScmObj
scm_disasm_dasm_inst_mod(ScmObj disasm, ScmObj mne)
{
  ScmObj mod = SCM_OBJ_INIT;

  SCM_REFSTK_INIT_REG(&mne,
                      &mod);

  scm_assert(scm_disassembler_p(disasm));
  scm_assert(scm_symbol_p(mne));

  if (scm_disasm_opd_mod__mod_id(disasm) >= 0) {
    mod = scm_make_number_from_size_t((size_t)scm_disasm_opd_mod__mod_id(disasm));
    if (scm_obj_null_p(mod)) return SCM_OBJ_NULL;
  }
  else {
    mod = scm_disasm_opd_mod__1(disasm);
  }

  return scm_list(2, mne, mod);
}

static ScmObj
scm_disasm_dasm_pinst_label(ScmObj disasm, ScmObj mne)
{
  ScmObj num = SCM_OBJ_INIT;

  SCM_REFSTK_INIT_REG(&mne,
                      &num);

  scm_assert(scm_disassembler_p(disasm));
  scm_assert(scm_symbol_p(mne));

  num = scm_make_number_from_size_t(scm_disasm_opd_label__id(disasm));
  if (scm_obj_null_p(num)) return SCM_OBJ_NULL;

  return scm_list(2, mne, num);
}

static ScmObj
scm_disasm_dasm_pinst_declmod(ScmObj disasm, ScmObj mne)
{
  ScmObj id = SCM_OBJ_INIT;

  SCM_REFSTK_INIT_REG(&mne,
                      &id);

  scm_assert(scm_disassembler_p(disasm));
  scm_assert(scm_symbol_p(mne));

  id = scm_make_number_from_size_t(scm_disasm_opd_declmod__id(disasm));
  if (scm_obj_null_p(id)) return SCM_OBJ_NULL;

  return scm_list(4, mne, id, scm_disasm_opd_declmod__mod(disasm),
                  (scm_disasm_opd_declmod__reg(disasm)
                   ? SCM_TRUE_OBJ :SCM_FALSE_OBJ));
}

static ScmObj
scm_disasm_dasm_pinst_sequence(ScmObj disasm, ScmObj mne)
{
  ScmObj id = SCM_OBJ_INIT;

  SCM_REFSTK_INIT_REG(&mne,
                      &id);

  scm_assert(scm_disassembler_p(disasm));
  scm_assert(scm_symbol_p(mne));

  id = scm_make_number_from_size_t(scm_disasm_opd_sequence__id(disasm));
  if (scm_obj_null_p(id)) return SCM_OBJ_NULL;

  return scm_list(2, mne, id);
}

static ScmObj
scm_disasm_dasm_inst(ScmObj disasm, ScmObj mne)
{
  scm_assert(scm_disassembler_p(disasm));
  scm_assert(scm_symbol_p(mne));

  switch (scm_disasm_fmt(disasm)) {
  case SCM_ASM_FMT_NOOPD:
    return scm_disasm_dasm_inst_noopd(disasm, mne);
    break;
  case SCM_ASM_FMT_OBJ:
    return scm_disasm_dasm_inst_obj(disasm, mne);
    break;
  case SCM_ASM_FMT_GV_MOD:
    return scm_disasm_dasm_inst_gv_mod(disasm, mne);
    break;
  case SCM_ASM_FMT_SI:
    return scm_disasm_dasm_inst_si(disasm, mne);
    break;
  case SCM_ASM_FMT_SI_SI:
    return scm_disasm_dasm_inst_si_si(disasm, mne);
    break;
  case SCM_ASM_FMT_SI_SI_SEQ:
    return scm_disasm_dasm_inst_si_si_seq(disasm, mne);
    break;
  case SCM_ASM_FMT_IOF:
    return scm_disasm_dasm_inst_iof(disasm, mne);
    break;
  case SCM_ASM_FMT_MOD:
    return scm_disasm_dasm_inst_mod(disasm, mne);
    break;
  case SCM_ASM_PI_LABEL:
    return scm_disasm_dasm_pinst_label(disasm, mne);
    break;
  case SCM_ASM_PI_DECLMOD:
    return scm_disasm_dasm_pinst_declmod(disasm, mne);
    break;
  case SCM_ASM_PI_SEQUENCE:
    return scm_disasm_dasm_pinst_sequence(disasm, mne);
    break;
  default:
    scm_error("failed to disassemble: unknown instructin format", 0);
    break;
  }

  return SCM_OBJ_NULL;
}

ScmObj
scm_disasm_disassemble(ScmObj disasm)
{
  ScmObj list = SCM_OBJ_INIT, tail = SCM_OBJ_INIT, inst = SCM_OBJ_INIT;
  ScmObj mne = SCM_OBJ_INIT;
  int r;

  SCM_REFSTK_INIT_REG(&disasm,
                      &list, &tail, &inst);

  scm_assert(scm_disassembler_p(disasm));

  list = tail = scm_cons(SCM_UNDEF_OBJ, SCM_NIL_OBJ);
  if (scm_obj_null_p(list)) return SCM_OBJ_NULL;

  while (!scm_disasm_terminate_p(disasm)) {
    mne = scm_asm_mnemonic(scm_disasm_op(disasm));
    if (scm_obj_null_p(mne)) return SCM_OBJ_NULL;

    inst = scm_disasm_dasm_inst(disasm, mne);
    if (scm_obj_null_p(inst)) return SCM_OBJ_NULL;

    inst = scm_cons(inst, SCM_NIL_OBJ);
    if (scm_obj_null_p(inst)) return SCM_OBJ_NULL;

    scm_set_cdr(tail, inst);
    tail = inst;

    r = scm_disasm_next(disasm);
    if (r < 0) return SCM_OBJ_NULL;
  }

  return scm_cdr(list);
}

static ScmObj
asm_get_asmb(ScmObj acc)
{
  if (scm_assembler_p(acc))
    return acc;
  else
    return scm_make_assembler(acc);
}

static ScmObj
asm_return_result(ScmObj asmb, ScmObj acc)
{
  if (scm_assembler_p(acc)) {
    return acc;
  }
  else {
    int r = scm_asm_commit(asmb);
    if (r < 0) return SCM_OBJ_NULL;

    return scm_asm_cdprogram(asmb);
  }
}

ScmObj
scm_assemble_1inst_cv(const ScmObj *inst, size_t n, ScmObj acc)
{
  ScmObj asmb = SCM_OBJ_INIT;
  int r;

  SCM_REFSTK_INIT_REG(&acc,
                      &asmb);

  scm_assert(n == 0 || inst != NULL);
  scm_assert(scm_obj_null_p(acc)
             || scm_iseq_p(acc) || scm_assembler_p(acc));

  asmb = asm_get_asmb(acc);
  if (scm_obj_null_p(asmb)) return SCM_OBJ_NULL;

  r = scm_asm_assemble_1inst_cv(asmb, inst, n);
  if (r < 0) return SCM_OBJ_NULL;

  return asm_return_result(asmb, acc);
}

ScmObj
scm_assemble_1inst(ScmObj inst, ScmObj acc)
{
  ScmObj asmb = SCM_OBJ_INIT;
  int r;

  SCM_REFSTK_INIT_REG(&inst, &acc,
                      &asmb);

  scm_assert(scm_pair_p(inst) || scm_nil_p(inst));
  scm_assert(scm_obj_null_p(acc)
             || scm_iseq_p(acc) || scm_assembler_p(acc));

  asmb = asm_get_asmb(acc);
  if (scm_obj_null_p(asmb)) return SCM_OBJ_NULL;

  r = scm_asm_assemble_1inst(asmb, inst);
  if (r < 0) return SCM_OBJ_NULL;

  return asm_return_result(asmb, acc);
}

ScmObj
scm_assemble(ScmObj lst, ScmObj acc)
{
  ScmObj asmb = SCM_OBJ_INIT;
  int r;

  SCM_REFSTK_INIT_REG(&lst, &acc,
                      &asmb);

  scm_assert(scm_pair_p(lst) || scm_nil_p(lst));
  scm_assert(scm_obj_null_p(acc)
             || scm_cdprogram_p(acc) || scm_assembler_p(acc));

  asmb = asm_get_asmb(acc);
  if (scm_obj_null_p(asmb)) return SCM_OBJ_NULL;

  r = scm_asm_assemble(asmb, lst);
  if (r < 0) return SCM_OBJ_NULL;

  return asm_return_result(asmb, acc);
}

ScmObj
scm_disassemble(ScmObj obj)
{
  scm_assert(scm_cdprogram_p(obj) || scm_disassembler_p(obj));

  if (scm_cdprogram_p(obj)) {
    obj = scm_make_disassembler(obj);
    if (scm_obj_null_p(obj)) return SCM_OBJ_NULL;
  }

  return scm_disasm_disassemble(obj);
}


/**************************************************************************/
/* Printable/Unprintable                                                  */
/**************************************************************************/

#define SCM_ASM_PRINTABLE_MNE_UNDEF      "undef"
#define SCM_ASM_PRINTABLE_MNE_UNINIT     "uninit"
#define SCM_ASM_PRINTABLE_MNE_QQTEMPLATE "qqtemplate"

static ScmObj
scm_asm_unprintable_inst_undef(ScmObj inst)
{
  ScmObj op = SCM_OBJ_INIT;

  op = scm_make_symbol_from_cstr(scm_asm_opcode2mnemonic(SCM_OPCODE_IMMVAL),
                                 SCM_ENC_SRC);
  if (scm_obj_null_p(op)) return SCM_OBJ_NULL;

  return scm_list(2, op, SCM_UNDEF_OBJ);
}

static ScmObj
scm_asm_unprintable_inst_uninit(ScmObj inst)
{
  ScmObj op = SCM_OBJ_INIT;

  op = scm_make_symbol_from_cstr(scm_asm_opcode2mnemonic(SCM_OPCODE_IMMVAL),
                                 SCM_ENC_SRC);
  if (scm_obj_null_p(op)) return SCM_OBJ_NULL;

  return scm_list(2, op, SCM_UNINIT_OBJ);
}

static ScmObj
scm_asm_unprintable_inst_qqtemplate(ScmObj inst)
{
  ScmObj op = SCM_OBJ_INIT, qq = SCM_OBJ_NULL;

  SCM_REFSTK_INIT_REG(&inst,
                      &op, &qq);

  op = scm_make_symbol_from_cstr(scm_asm_opcode2mnemonic(SCM_OPCODE_IMMVAL),
                                 SCM_ENC_SRC);
  if (scm_obj_null_p(op)) return SCM_OBJ_NULL;

  qq = scm_list_ref(inst, 1);
  if (scm_obj_null_p(qq)) return SCM_OBJ_NULL;

  if (!scm_qqtmpl_p(qq)) {
    qq = scm_compile_qq_template(qq);
    if (scm_obj_null_p(qq)) return SCM_OBJ_NULL;
  }

  return scm_list(2, op, qq);
}

static ScmObj
scm_asm_printable_inst_immval(ScmObj inst)
{
  ScmObj val = SCM_OBJ_INIT, op = SCM_OBJ_INIT;

  SCM_REFSTK_INIT_REG(&inst,
                      &val, &op);

  val = scm_list_ref(inst, 1);
  if (scm_obj_null_p(val)) return SCM_OBJ_NULL;

  if (scm_undef_object_p(val)) {
    op = scm_make_symbol_from_cstr(SCM_ASM_PRINTABLE_MNE_UNDEF, SCM_ENC_SRC);
    if (scm_obj_null_p(op)) return SCM_OBJ_NULL;
    return scm_list(1, op);
  }
  else if (scm_landmine_object_p(val)) {
    op = scm_make_symbol_from_cstr(SCM_ASM_PRINTABLE_MNE_UNINIT, SCM_ENC_SRC);
    if (scm_obj_null_p(op)) return SCM_OBJ_NULL;
    return scm_list(1, op);
  }
  else if (scm_qqtmpl_p(val)) {
    op = scm_make_symbol_from_cstr(SCM_ASM_PRINTABLE_MNE_QQTEMPLATE,
                                   SCM_ENC_SRC);
    if (scm_obj_null_p(op)) return SCM_OBJ_NULL;

    val = scm_qqtmpl_template(val);
    if (scm_obj_null_p(val)) return SCM_OBJ_NULL;

    return scm_list(2, op, val);
  }
  else {
    return inst;
  }
}

static ScmObj
scm_asm_printable_inst_global_var(ScmObj inst)
{
  ScmObj module = SCM_OBJ_INIT, var = SCM_OBJ_INIT;

  SCM_REFSTK_INIT_REG(&inst,
                      &module, &var);

  var = scm_list_ref(inst, 1);
  if (scm_obj_null_p(var)) return SCM_OBJ_NULL;

  module = scm_list_ref(inst, 2);
  if (scm_obj_null_p(module)) return SCM_OBJ_NULL;

  if (scm_gloc_p(var))
    var = scm_gloc_symbol(var);

  if (scm_module_p(module))
    module = scm_module_name(module);

  return scm_list(3, scm_car(inst), var, module);
}

struct scm_asm_cnv {
  const char *op;
  ScmObj (*func)(ScmObj);
};

static ScmObj
scm_asm_printable_inst_module(ScmObj inst)
{
  ScmObj module = SCM_OBJ_INIT;

  SCM_REFSTK_INIT_REG(&inst,
                      &module);

  module = scm_list_ref(inst, 1);
  if (scm_obj_null_p(module)) return SCM_OBJ_NULL;

  if (scm_module_p(module))
    module = scm_module_name(module);

  return scm_list(2, scm_car(inst), module);
}

static ScmObj
scm_asm_printable_pinst_declmod(ScmObj inst)
{
  ScmObj module = SCM_OBJ_INIT, printable = SCM_OBJ_INIT;
  int r;

  SCM_REFSTK_INIT_REG(&inst,
                      &module, &printable);

  module = scm_list_ref(inst, 2);
  if (scm_obj_null_p(module)) return SCM_OBJ_NULL;

  if (scm_module_p(module))
    module = scm_module_name(module);

  printable = scm_list_copy(inst);
  if (scm_obj_null_p(printable)) return SCM_OBJ_NULL;

  r = scm_list_set(printable, 2, module);
  if (r < 0) return SCM_OBJ_NULL;

  return printable;
}

static ScmObj
scm_asm_convert_inst(ScmObj inst, const struct scm_asm_cnv *cnv)
{
  ScmObj op = SCM_OBJ_INIT;
  bool cmp;
  int r;

  SCM_REFSTK_INIT_REG(&inst,
                      &op);

  scm_assert(scm_pair_p(inst));

  op = scm_car(inst);
  for (size_t i = 0; cnv[i].op != NULL; i++) {
    r = scm_symbol_eq_cstr(op, cnv[i].op, SCM_ENC_SRC, &cmp);
    if (r < 0) return SCM_OBJ_NULL;

    if (cmp)
      return cnv[i].func(inst);
  }

  return inst;
}

ScmObj
scm_unprintable_asm_inst(ScmObj inst)
{
  static const struct scm_asm_cnv tbl[] = {
    { .op   = SCM_ASM_PRINTABLE_MNE_UNDEF,
      .func = scm_asm_unprintable_inst_undef },
    { .op   = SCM_ASM_PRINTABLE_MNE_UNINIT,
      .func = scm_asm_unprintable_inst_uninit },
    { .op   = SCM_ASM_PRINTABLE_MNE_QQTEMPLATE,
      .func = scm_asm_unprintable_inst_qqtemplate },
    { .op   = NULL,
      .func = NULL },
  };

  scm_assert(scm_pair_p(inst));
  return scm_asm_convert_inst(inst, tbl);
}

ScmObj
scm_printable_asm_inst(ScmObj inst)
{
  static const struct scm_asm_cnv tbl[] = {
    { .op = "immval",  .func = scm_asm_printable_inst_immval, },
    { .op = "gdef",    .func = scm_asm_printable_inst_global_var },
    { .op = "gref",    .func = scm_asm_printable_inst_global_var },
    { .op = "gset",    .func = scm_asm_printable_inst_global_var },
    { .op = "module",  .func = scm_asm_printable_inst_module },
    { .op = "declmod", .func = scm_asm_printable_pinst_declmod },
    { .op = NULL,      .func = NULL },
  };

  scm_assert(scm_pair_p(inst));
  return scm_asm_convert_inst(inst, tbl);
}

ScmObj
scm_unprintable_asm(ScmObj lst)
{
  ScmObj inst = SCM_OBJ_INIT, rest = SCM_OBJ_INIT;

  SCM_REFSTK_INIT_REG(&lst,
                      &inst);

  if (!scm_pair_p(lst))
    return lst;

  inst = scm_unprintable_asm_inst(scm_car(lst));
  if (scm_obj_null_p(inst)) return SCM_OBJ_NULL;

  rest = scm_unprintable_asm(scm_cdr(lst));
  if (scm_obj_null_p(rest)) return SCM_OBJ_NULL;

  return scm_cons(inst, rest);
}

ScmObj
scm_printable_asm(ScmObj lst)
{
  ScmObj inst = SCM_OBJ_INIT, rest = SCM_OBJ_INIT;

  SCM_REFSTK_INIT_REG(&lst,
                      &inst);

  if (!scm_pair_p(lst))
    return lst;

  inst = scm_printable_asm_inst(scm_car(lst));
  if (scm_obj_null_p(inst)) return SCM_OBJ_NULL;

  rest = scm_printable_asm(scm_cdr(lst));
  if (scm_obj_null_p(rest)) return SCM_OBJ_NULL;

  return scm_cons(inst, rest);
}


/**************************************************************************/
/* etc                                                                    */
/**************************************************************************/

int
scm_asm_mnemonic2opcode(const char *mne)
{
  scm_assert(mne != NULL);

  for (size_t i = 0;
       i < sizeof(opcode2mnemonic_tbl)/sizeof(opcode2mnemonic_tbl[0]);
       i++) {
    if (strcmp(opcode2mnemonic_tbl[i].mne, mne) == 0)
      return (int)opcode2mnemonic_tbl[i].code;
  }

  return -1;
}

const char *
scm_asm_opcode2mnemonic(scm_opcode_t code)
{
  for (size_t i = 0;
       i < sizeof(opcode2mnemonic_tbl)/sizeof(opcode2mnemonic_tbl[0]);
       i++) {
    if (opcode2mnemonic_tbl[i].code == code)
      return opcode2mnemonic_tbl[i].mne;
  }

  return NULL;
}

ScmObj
scm_asm_mnemonic(scm_opcode_t opcode)
{
  const char *p;

  p = scm_asm_opcode2mnemonic(opcode);
  if (p == NULL) {
    scm_error("assembler: unknown opcode", 0);
    return SCM_OBJ_NULL;
  }

  return  scm_make_symbol_from_cstr(p, SCM_ENC_SRC);
}

void
scm_asm_update_iof(ScmObj iseq, scm_byte_t *ip, int iof)
{
  scm_assert(scm_iseq_p(iseq));
  scm_assert(scm_iseq_to_ip(iseq) + SCM_VM_INST_OFFSET_OPD <= ip
             && ip < (scm_iseq_to_ip(iseq)
                      + (ptrdiff_t)scm_iseq_length(iseq)
                      - SCM_VM_INST_SZ_OPD_IOF));

  SCM_WB_EXP(iseq,
             scm_vm_inst_update_opd_iof(ip - SCM_VM_INST_OFFSET_OPD,
                                       iof, SCM_VM_INST_UPD_FLG_OPD1));
}

int
scm_asm_replace_symbol_with_gloc(ScmObj clsr, scm_byte_t *ip, ScmObj gloc)
{
  ScmObj iseq = SCM_OBJ_INIT;

  SCM_REFSTK_INIT_REG(&clsr, &gloc,
                      &iseq);

  scm_assert(scm_closure_p(clsr) || scm_cdprogram_p(clsr));
  scm_assert(scm_gloc_p(gloc));

  if (scm_closure_p(clsr))
    iseq = scm_closure_iseq(clsr);
  else
    iseq = scm_cdprogram_head_iseq(clsr);

  if (scm_obj_null_p(iseq)) return -1;

  scm_assert(scm_iseq_to_ip(iseq) + SCM_VM_INST_OFFSET_OPD <= ip
             && ip < (scm_iseq_to_ip(iseq)
                      + (ptrdiff_t)scm_iseq_length(iseq)
                      - SCM_VM_INST_SZ_OPD_OBJ_OBJ));

  SCM_WB_EXP(iseq,
             scm_vm_inst_update_opd_obj_obj(ip - SCM_VM_INST_OFFSET_OPD,
                                            gloc, SCM_OBJ_NULL,
                                            SCM_VM_INST_UPD_FLG_OPD1));

  return 0;
}
