#ifndef INCLUDE_ASSEMBLER_H__
#define INCLUDE_ASSEMBLER_H__

#include <sys/types.h>
#include <stddef.h>
#include <stdint.h>
#include <stdbool.h>

#include "scythe/object.h"
#include "scythe/vminst.h"
#include "scythe/earray.h"
#include "scythe/memory.h"

#define SCM_ASM_PI_CODE_START ((INT_MAX >> 1) + 1)

enum {
  SCM_ASM_PI_LABEL = SCM_ASM_PI_CODE_START,  /* define a label */
  SCM_ASM_PI_DECLMOD,
  SCM_ASM_PI_SEQUENCE,

  SCM_ASM_PI_CODE_END
};

#define SCM_ASM_NR_PI (SCM_ASM_PI_CODE_END - SCM_ASM_PI_CODE_START)
#define SCM_ASM_NR_OPD_MAX 3


/***************************************************************************/
/*  Assembler Instruction Format                                           */
/***************************************************************************/

enum {
  SCM_ASM_FMT_NOOPD,
  SCM_ASM_FMT_OBJ,
  SCM_ASM_FMT_GV_MOD,
  SCM_ASM_FMT_SI,
  SCM_ASM_FMT_SI_SI,
  SCM_ASM_FMT_SI_SI_SEQ,
  SCM_ASM_FMT_IOF,
  SCM_ASM_FMT_MOD,
};

int scm_asm_inst_format(scm_opcode_t op);


/***************************************************************************/
/*  AsmModDelcs                                                            */
/***************************************************************************/

typedef struct ScmAsmModDeclsItrRec ScmAsmModDeclsItr;
typedef struct ScmAsmModDeclElmRec ScmAsmModDeclElm;
typedef struct ScmAsmModDeclsRec ScmAsmModDecls;

struct ScmAsmModDeclsItrRec {
  EArray indexes;
  size_t cur;
};

struct ScmAsmModDeclElmRec {
  size_t seq_id;
  size_t offset;
  ScmObj module;
  bool registered;
};

struct ScmAsmModDeclsRec {
  ScmObjHeader header;
  EArray decls;
};

#define SCM_ASMMODDECLS(obj) ((ScmAsmModDecls *)(obj))
#define SCM_ASMMODDECLS_DECLS(c) (&SCM_ASMMODDECLS(c)->decls)

extern ScmTypeInfo SCM_ASMMODDECLS_TYPE_INFO;

int scm_asmmoddecls_initialize(ScmObj asml);
void scm_asmmoddecls_finalize(ScmObj obj);
ScmObj scm_asmmoddecls_new(scm_mem_type_t mtype);
ScmObj scm_asmmoddecls_module(ScmObj asml, size_t id);
int scm_asmmoddecls_registered_flag(ScmObj asml, size_t id, bool *flg);
int scm_asmmoddecls_id(ScmObj asml, ScmObj mod, ssize_t *id);
ssize_t scm_asmmoddecls_assign_id(ScmObj asml);
int scm_asmmoddecls_register_id(ScmObj asml, size_t id);
int scm_asmmoddecls_register_decl(ScmObj asml, size_t id,
                                  size_t seq_id, size_t offset,
                                  ScmObj mod, bool reg);
void scm_asmmoddecls_itr_init(ScmAsmModDeclsItr *itr);
void scm_asmmoddecls_itr_fin(ScmAsmModDeclsItr *itr);
int scm_asmmoddecls_itr_begin(ScmObj asml, size_t seq_id,
                              ScmAsmModDeclsItr *itr);
bool scm_asmmoddecls_itr_end_p(ScmObj asml, ScmAsmModDeclsItr *itr);
void scm_asmmoddecls_itr_next(ScmObj asml, ScmAsmModDeclsItr *itr);
size_t scm_asmmoddecls_itr_id(ScmObj asml, ScmAsmModDeclsItr *itr);
size_t scm_asmmoddecls_itr_offset(ScmObj asml, ScmAsmModDeclsItr *itr);
ScmObj scm_asmmoddecls_itr_module(ScmObj asml, ScmAsmModDeclsItr *itr);
bool scm_asmmoddecls_itr_registered(ScmObj asml, ScmAsmModDeclsItr *itr);
void scm_asmmoddecls_gc_initialize(ScmObj obj);
void scm_asmmoddecls_gc_finalize(ScmObj obj);
int scm_asmmoddecls_gc_accept(ScmObj obj, ScmGCRefHandler handler);

static inline bool
scm_asmmoddecls_p(ScmObj obj)
{
  return scm_obj_type_p(obj, &SCM_ASMMODDECLS_TYPE_INFO);
}

static inline bool
scm_asmmoddecls_assigned_p(ScmObj asml, size_t id)
{
  scm_assert_obj_type(asml, &SCM_ASMMODDECLS_TYPE_INFO);
  return (id <= EARY_SIZE(SCM_ASMMODDECLS_DECLS(asml)));
}

static inline bool
scm_asmmoddecls_declared_p(ScmObj asml, size_t id)
{
  const ScmAsmModDeclElm *d;

  scm_assert_obj_type(asml, &SCM_ASMMODDECLS_TYPE_INFO);

  if (!scm_asmmoddecls_assigned_p(asml, id))
    return false;

  d = eary_idx_to_ptr(SCM_ASMMODDECLS_DECLS(asml), id);
  return scm_obj_not_null_p(d->module);
}

static inline size_t
scm_asmmoddecls_size(ScmObj asml)
{
  scm_assert_obj_type(asml, &SCM_ASMMODDECLS_TYPE_INFO);
  return EARY_SIZE(SCM_ASMMODDECLS_DECLS(asml));
}

static inline void
scm_asmmoddecls_clear(ScmObj asml)
{
  scm_assert_obj_type(asml, &SCM_ASMMODDECLS_TYPE_INFO);
  eary_truncate(SCM_ASMMODDECLS_DECLS(asml));
}

static inline ScmObj
scm_make_asmmoddecls(void)
{
  return scm_asmmoddecls_new(SCM_MEM_HEAP);
}


/**************************************************************************/
/* Assembler                                                              */
/**************************************************************************/

typedef struct ScmAsmLabelDeclRec ScmAsmLabelDecl;
typedef struct ScmAsmLabelRefRec ScmAsmLabelRef;
typedef struct ScmAssemblerRec ScmAssembler;

struct ScmAsmLabelDeclRec {
  ssize_t offset;
};

struct ScmAsmLabelRefRec{
  size_t offset;
  size_t label_id;
};

struct ScmAssemblerRec {
  ScmObjHeader header;
  size_t id;
  ScmObj prog;
  ScmObj iseq;
  EArray label_decl;
  EArray label_ref;
};

#define SCM_ASSEMBLER(obj) ((ScmAssembler *)(obj))
#define SCM_ASSEMBLER_ID(a) (SCM_ASSEMBLER(a)->id)
#define SCM_ASSEMBLER_PROG(a) (SCM_ASSEMBLER(a)->prog)
#define SCM_ASSEMBLER_ISEQ(a) (SCM_ASSEMBLER(a)->iseq)
#define SCM_ASSEMBLER_LABEL_DECL(a) (&SCM_ASSEMBLER(a)->label_decl)
#define SCM_ASSEMBLER_LABEL_REF(a) (&SCM_ASSEMBLER(a)->label_ref)

extern ScmTypeInfo SCM_ASSEMBLER_TYPE_INFO;

int scm_asm_initialize(ScmObj asmb, ScmObj prog);
void scm_asm_finalize(ScmObj obj);
ScmObj scm_asm_new(scm_mem_type_t mtype, ScmObj prog);
ScmObj scm_asm_new_seq(ScmObj asmb, size_t id);
ScmObj scm_asm_head_iseq(ScmObj asmb);
ssize_t scm_asm_assign_label_id(ScmObj asmb);
int scm_asm_register_label_id(ScmObj asmb, size_t id);
ssize_t scm_asm_assign_module_id(ScmObj asmb);
int scm_asm_register_module_id(ScmObj asmb, size_t id);
ssize_t scm_asm_assign_sequence_id(ScmObj asmb);
int scm_asm_register_sequence_id(ScmObj asmb, size_t id);
int scm_asm_push_inst_noopd(ScmObj asmb, scm_opcode_t op);
int scm_asm_push_inst_obj(ScmObj asmb, scm_opcode_t op, ScmObj obj);
int scm_asm_push_inst_gv_mod(ScmObj asmb,
                             scm_opcode_t op, ScmObj gv, ScmObj mod);
int scm_asm_push_inst_si(ScmObj asmb, scm_opcode_t op, int si);
int scm_asm_push_inst_si_si(ScmObj asmb, scm_opcode_t op, int si1, int si2);
int scm_asm_push_inst_si_si_seq(ScmObj asmb,
                                scm_opcode_t op, int si1, int si2, ScmObj seq);
int scm_asm_push_inst_iof(ScmObj asmb, scm_opcode_t op, bool label, ...);
int scm_asm_push_inst_rlid(ScmObj asmb, scm_opcode_t op, size_t id);
int scm_asm_push_pinst_label(ScmObj asmb, scm_opcode_t op, size_t id);
int scm_asm_push_pinst_declmod(ScmObj asmb, scm_opcode_t op,
                               size_t id, ScmObj mod, bool reg);
int scm_asm_push_pinst_sequence(ScmObj asmb, scm_opcode_t op, size_t id);
int scm_asm_push_va(ScmObj asmb, scm_opcode_t op, va_list operands);
int scm_asm_push(ScmObj asmb, scm_opcode_t op, ...);
int scm_asm_resolve_label_ref(ScmObj asmb);
void scm_asm_clear_labels(ScmObj asmb);
int scm_asm_commit(ScmObj asmb);
void scm_asm_gc_initialize(ScmObj obj);
void scm_asm_gc_finalize(ScmObj obj);
int scm_asm_gc_accept(ScmObj obj, ScmGCRefHandler handler);

static inline bool
scm_assembler_p(ScmObj obj)
{
  return scm_obj_type_p(obj, &SCM_ASSEMBLER_TYPE_INFO);
}

static inline bool
scm_asm_uncommited_p(ScmObj asmb)
{
  scm_assert_obj_type(asmb, &SCM_ASSEMBLER_TYPE_INFO);
  return (EARY_SIZE(SCM_ASSEMBLER_LABEL_DECL(asmb)) > 0
          || EARY_SIZE(SCM_ASSEMBLER_LABEL_REF(asmb)) > 0);
}

static inline ScmObj
scm_asm_cdprogram(ScmObj asmb)
{
  scm_assert_obj_type(asmb, &SCM_ASSEMBLER_TYPE_INFO);
  return SCM_ASSEMBLER_PROG(asmb);
}

static inline ScmObj
scm_asm_iseq(ScmObj asmb)
{
  scm_assert_obj_type(asmb, &SCM_ASSEMBLER_TYPE_INFO);
  return SCM_ASSEMBLER_ISEQ(asmb);
}

static inline ScmObj
scm_make_assembler(ScmObj prog)
{
  return scm_asm_new(SCM_MEM_HEAP, prog);
}


/**************************************************************************/
/* Disassembler                                                           */
/**************************************************************************/

enum {
  SCM_DISASM_TK_INST,
  SCM_DISASM_TK_PSEUDO,
  SCM_DISASM_TK_END,
};

typedef struct ScmDisasmTokenRec ScmDisasmToken;
typedef struct ScmDisassemblerRec ScmDisassembler;

struct ScmDisasmTokenRec {
  int type;
  int fmt;
  scm_opcode_t op;
  union {
    struct { struct scm_vm_inst_opd_noopd o; } noopd;
    struct { struct scm_vm_inst_opd_obj o; ssize_t mod_id; } obj;
    struct { struct scm_vm_inst_opd_obj_obj o; ssize_t mod_id; } obj_obj;
    struct { struct scm_vm_inst_opd_si o; } si;
    struct { struct scm_vm_inst_opd_si_si o; } si_si;
    struct { struct scm_vm_inst_opd_si_si_obj o; size_t seq_id; } si_si_obj;
    struct { struct scm_vm_inst_opd_iof o; size_t label_id; } iof;
    struct { size_t id; } label;
    struct { size_t id; ScmObj mod; bool reg; } declmod;
    struct { size_t id; } sequence;
  } opd;
};

struct ScmDisassemblerRec {
  ScmObjHeader header;
  ScmObj prog;
  ScmObj iseq;
  ScmObj mod_decls;
  EArray label_decls;
  size_t seq_id;
  scm_byte_t *ip;
  size_t decl_label_idx;
  ScmAsmModDeclsItr decl_mod_itr;
  ScmDisasmToken token;
};

#define SCM_DISASSEMBLER(obj) ((ScmDisassembler *)(obj))
#define SCM_DISASSEMBLER_PROG(d) (SCM_DISASSEMBLER(d)->prog)
#define SCM_DISASSEMBLER_ISEQ(d) (SCM_DISASSEMBLER(d)->iseq)
#define SCM_DISASSEMBLER_MOD_DECLS(d) (SCM_DISASSEMBLER(d)->mod_decls)
#define SCM_DISASSEMBLER_LABEL_DECLS(d) (&SCM_DISASSEMBLER(d)->label_decls)
#define SCM_DISASSEMBLER_SEQ_ID(d) (SCM_DISASSEMBLER(d)->seq_id)
#define SCM_DISASSEMBLER_IP(d) (SCM_DISASSEMBLER(d)->ip)
#define SCM_DISASSEMBLER_DECL_LABEL_IDX(d) (SCM_DISASSEMBLER(d)->decl_label_idx)
#define SCM_DISASSEMBLER_DECL_MOD_ITR(d) (&SCM_DISASSEMBLER(d)->decl_mod_itr)
#define SCM_DISASSEMBLER_TOKEN(d) (&SCM_DISASSEMBLER(d)->token)

#define SCM_DISASSEMBLER_SET_ISEQ(d, v) SCM_SLOT_SETQ(ScmDisassembler, d, iseq, v)
#define SCM_DISASSEMBLER_SET_SEQ_ID(d, v) (SCM_DISASSEMBLER(d)->seq_id = (v))
#define SCM_DISASSEMBLER_SET_IP(d, v) (SCM_DISASSEMBLER(d)->ip = (v))
#define SCM_DISASSEMBLER_ADD_IP(d, a) (SCM_DISASSEMBLER(d)->ip += (a))
#define SCM_DISASSEMBLER_SET_DECL_LABEL_IDX(d, v) \
  (SCM_DISASSEMBLER(d)->decl_label_idx = (v))
#define SCM_DISASSEMBLER_INC_DECL_LABEL_IDX(d) \
  (SCM_DISASSEMBLER(d)->decl_label_idx++)
#define SCM_DISASSEMBLER_SET_DECL_MODULE_IDX(d, v) \
  (SCM_DISASSEMBLER(d)->decl_module_idx = (v))
#define SCM_DISASSEMBLER_INC_DECL_MODULE_IDX(d) \
  (SCM_DISASSEMBLER(d)->decl_module_idx++)

extern ScmTypeInfo SCM_DISASSEMBLER_TYPE_INFO;

int scm_disasm_initialize(ScmObj disasm, ScmObj prog);
void scm_disasm_finalize(ScmObj disasm);
ScmObj scm_disasm_new(scm_mem_type_t mtype, ScmObj prog);
int scm_disasm_next(ScmObj disasm);
void scm_disasm_rewind(ScmObj disasm);
void scm_disasm_gc_initialize(ScmObj obj);
void scm_disasm_gc_finalize(ScmObj obj);
int scm_disasm_gc_accept(ScmObj obj, ScmGCRefHandler handler);

static inline bool
scm_disassembler_p(ScmObj obj)
{
  return scm_obj_type_p(obj, &SCM_DISASSEMBLER_TYPE_INFO);
}

static inline ScmObj
scm_make_disassembler(ScmObj prog)
{
  return scm_disasm_new(SCM_MEM_HEAP, prog);
}

static inline bool
scm_disasm_terminate_p(ScmObj disasm)
{
  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);
  return (SCM_DISASSEMBLER_TOKEN(disasm)->type == SCM_DISASM_TK_END);
}

static inline int
scm_disasm_fmt(ScmObj disasm)
{
  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);
  return SCM_DISASSEMBLER_TOKEN(disasm)->fmt;
}

static inline scm_opcode_t
scm_disasm_op(ScmObj disasm)
{
  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);
  return SCM_DISASSEMBLER_TOKEN(disasm)->op;
}

static inline ScmObj
scm_disasm_opd_obj__1(ScmObj disasm)
{
  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);
  return SCM_DISASSEMBLER_TOKEN(disasm)->opd.obj.o.o1;
}

static inline ScmObj
scm_disasm_opd_gv_mod__1(ScmObj disasm)
{
  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);
  return SCM_DISASSEMBLER_TOKEN(disasm)->opd.obj_obj.o.o1;
}

static inline ScmObj
scm_disasm_opd_gv_mod__2(ScmObj disasm)
{
  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);
  return SCM_DISASSEMBLER_TOKEN(disasm)->opd.obj_obj.o.o2;
}

static inline ssize_t
scm_disasm_opd_gv_mod__mod_id(ScmObj disasm)
{
  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);
  return SCM_DISASSEMBLER_TOKEN(disasm)->opd.obj_obj.mod_id;
}

static inline int
scm_disasm_opd_si__1(ScmObj disasm)
{
  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);
  return SCM_DISASSEMBLER_TOKEN(disasm)->opd.si.o.o1;
}

static inline int
scm_disasm_opd_si_si__1(ScmObj disasm)
{
  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);
  return SCM_DISASSEMBLER_TOKEN(disasm)->opd.si_si.o.o1;
}

static inline int
scm_disasm_opd_si_si__2(ScmObj disasm)
{
  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);
  return SCM_DISASSEMBLER_TOKEN(disasm)->opd.si_si.o.o2;
}

static inline int
scm_disasm_opd_si_si_seq__1(ScmObj disasm)
{
  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);
  return SCM_DISASSEMBLER_TOKEN(disasm)->opd.si_si_obj.o.o1;
}

static inline int
scm_disasm_opd_si_si_seq__2(ScmObj disasm)
{
  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);
  return SCM_DISASSEMBLER_TOKEN(disasm)->opd.si_si_obj.o.o2;
}

static inline ScmObj
scm_disasm_opd_si_si_seq__3(ScmObj disasm)
{
  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);
  return SCM_DISASSEMBLER_TOKEN(disasm)->opd.si_si_obj.o.o3;
}

static inline size_t
scm_disasm_opd_si_si_seq__seq_id(ScmObj disasm)
{
  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);
  return SCM_DISASSEMBLER_TOKEN(disasm)->opd.si_si_obj.seq_id;
}

static inline int
scm_disasm_opd_iof__1(ScmObj disasm)
{
  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);
  return SCM_DISASSEMBLER_TOKEN(disasm)->opd.iof.o.o1;
}

static inline size_t
scm_disasm_opd_iof__label_id(ScmObj disasm)
{
  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);
  return SCM_DISASSEMBLER_TOKEN(disasm)->opd.iof.label_id;
}

static inline ScmObj
scm_disasm_opd_mod__1(ScmObj disasm)
{
  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);
  return SCM_DISASSEMBLER_TOKEN(disasm)->opd.obj.o.o1;
}

static inline ssize_t
scm_disasm_opd_mod__mod_id(ScmObj disasm)
{
  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);
  return SCM_DISASSEMBLER_TOKEN(disasm)->opd.obj.mod_id;
}

static inline size_t
scm_disasm_opd_label__id(ScmObj disasm)
{
  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);
  return SCM_DISASSEMBLER_TOKEN(disasm)->opd.label.id;
}

static inline size_t
scm_disasm_opd_declmod__id(ScmObj disasm)
{
  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);
  return SCM_DISASSEMBLER_TOKEN(disasm)->opd.declmod.id;
}

static inline ScmObj
scm_disasm_opd_declmod__mod(ScmObj disasm)
{
  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);
  return SCM_DISASSEMBLER_TOKEN(disasm)->opd.declmod.mod;
}

static inline bool
scm_disasm_opd_declmod__reg(ScmObj disasm)
{
  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);
  return SCM_DISASSEMBLER_TOKEN(disasm)->opd.declmod.reg;
}

static inline size_t
scm_disasm_opd_sequence__id(ScmObj disasm)
{
  scm_assert_obj_type(disasm, &SCM_DISASSEMBLER_TYPE_INFO);
  return SCM_DISASSEMBLER_TOKEN(disasm)->opd.sequence.id;
}


/**************************************************************************/
/* Assemble/Disassemble                                                   */
/**************************************************************************/

int scm_asm_assemble_1inst_cv(ScmObj asmb, const ScmObj *inst, size_t n);
int scm_asm_assemble_1inst(ScmObj asmb, ScmObj inst);
int scm_asm_assemble(ScmObj asmb, ScmObj lst);
ScmObj scm_disasm_disassemble(ScmObj disasm);

ScmObj scm_assemble_1inst_cv(const ScmObj *inst, size_t n, ScmObj acc);
ScmObj scm_assemble_1inst(ScmObj inst, ScmObj acc);
ScmObj scm_assemble(ScmObj lst, ScmObj acc);
ScmObj scm_disassemble(ScmObj obj);


/**************************************************************************/
/* Printable/Unprintable                                                  */
/**************************************************************************/

ScmObj scm_unprintable_asm_inst(ScmObj inst);
ScmObj scm_printable_asm_inst(ScmObj inst);
ScmObj scm_unprintable_asm(ScmObj lst);
ScmObj scm_printable_asm(ScmObj lst);


/**************************************************************************/
/* etc                                                                    */
/**************************************************************************/

int scm_asm_mnemonic2opcode(const char *mne);
const char *scm_asm_opcode2mnemonic(scm_opcode_t code);
ScmObj scm_asm_mnemonic(scm_opcode_t opcode);
void scm_asm_update_iof(ScmObj iseq, scm_byte_t *ip, int iof);
int scm_asm_replace_symbol_with_gloc(ScmObj clsr, scm_byte_t *ip, ScmObj gloc);


#endif /* INCLUDE_ASSEMBLER_H__ */
