#include <sys/types.h>
#include <stddef.h>
#include <stdbool.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <assert.h>

#include "scythe/object.h"
#include "scythe/bedrock.h"
#include "scythe/memory.h"
#include "scythe/refstk.h"
#include "scythe/exception.h"
#include "scythe/earray.h"
#include "scythe/impl_utils.h"
#include "scythe/iseq.h"


/***************************************************************************/
/*  ISeq                                                                   */
/***************************************************************************/

ScmTypeInfo SCM_ISEQ_TYPE_INFO = {
  .name                = "iseq",
  .flags               = SCM_TYPE_FLG_MMO,
  .obj_print_func      = NULL,
  .obj_size            = sizeof(ScmISeq),
  .gc_ini_func         = scm_iseq_gc_initialize,
  .gc_fin_func         = scm_iseq_gc_finalize,
  .gc_accept_func      = scm_iseq_gc_accept,
  .gc_accept_func_weak = NULL,
  .extra               = NULL,
};

int
scm_iseq_initialize(ScmObj iseq)
{
  int rslt;

  scm_assert_obj_type(iseq, &SCM_ISEQ_TYPE_INFO);

  rslt = eary_init(SCM_ISEQ_EARY_SEQ(iseq),
                   sizeof(scm_byte_t), SCM_ISEQ_DEFAULT_SEQ_SIZE);
  if (rslt != 0) return -1;

  rslt = eary_init(SCM_ISEQ_EARY_OBJS(iseq),
                   sizeof(size_t), SCM_ISEQ_DEFAULT_OBJS_SIZE);
  if (rslt != 0) return -1;

  return 0;
}

void
scm_iseq_finalize(ScmObj obj)
{
  scm_assert_obj_type(obj, &SCM_ISEQ_TYPE_INFO);

  eary_fin(SCM_ISEQ_EARY_SEQ(obj));
  eary_fin(SCM_ISEQ_EARY_OBJS(obj));
}

ScmObj
scm_iseq_new(scm_mem_type_t mtype)
{
  ScmObj iseq = SCM_OBJ_INIT;

  SCM_REFSTK_INIT_REG(&iseq);

  iseq = scm_alloc_mem(&SCM_ISEQ_TYPE_INFO, 0, mtype);
  if (scm_obj_null_p(iseq)) return SCM_OBJ_NULL;

  if (scm_iseq_initialize(iseq) < 0)
    return SCM_OBJ_NULL;

  return iseq;
}

ssize_t
scm_iseq_ip_to_offset(ScmObj iseq, scm_byte_t *ip)
{
  size_t offset;

  scm_assert_obj_type(iseq, &SCM_ISEQ_TYPE_INFO);

  if (ip < SCM_ISEQ_SEQ_VEC(iseq)) {
    scm_error("failed to calculate offset of VM instruction: out of range",
                  0);
    return -1;
  }

  offset = (size_t)(ip - SCM_ISEQ_SEQ_VEC(iseq));
  if (offset > SCM_ISEQ_SEQ_LENGTH(iseq)) {
    scm_error("failed to calculate offset of VM instruction: out of range",
                  0);
    return -1;
  }
  else if (offset > SSIZE_MAX) {
    scm_error("failed to calculate offset of VM instruction: "
                  "offset value too big", 0);
    return -1;
  }

  return (ssize_t)offset;
}

ssize_t
scm_iseq_push_inst(ScmObj iseq, const void *inst, size_t sz,
                   const size_t *objs, size_t n)
{
  size_t offset;
  scm_byte_t *ip;
  int r;

  scm_assert_obj_type(iseq, &SCM_ISEQ_TYPE_INFO);
  scm_assert(inst != NULL);
  scm_assert(n == 0 || objs != NULL);

  if (SCM_ISEQ_SEQ_LENGTH(iseq) > SSIZE_MAX - sz) {
    scm_error("failed to expand ISeq: overflow", 0);
    return -1;
  }

  offset = SCM_ISEQ_SEQ_LENGTH(iseq);
  EARY_SET(SCM_ISEQ_EARY_SEQ(iseq), scm_byte_t, offset + sz - 1, 0, r);
  if (r < 0) return -1;

  ip = scm_iseq_to_ip(iseq) + offset;
  if (objs != NULL) {
    for (size_t i = 0; i < n; i++) {
      EARY_PUSH(SCM_ISEQ_EARY_OBJS(iseq), size_t, offset + objs[i], r);
      if(r < 0) return -1;
    }

    SCM_WB_EXP(iseq, memcpy(ip, inst, sz));
  }
  else {
    memcpy(ip, inst, sz);
  }

  return (ssize_t)(offset + sz);
}

void
scm_iseq_gc_initialize(ScmObj obj)
{
  scm_assert_obj_type(obj, &SCM_ISEQ_TYPE_INFO);

  eary_init(SCM_ISEQ_EARY_SEQ(obj), 0, 0);
  eary_init(SCM_ISEQ_EARY_OBJS(obj), 0, 0);
}

void
scm_iseq_gc_finalize(ScmObj obj)
{
  scm_iseq_finalize(obj);
}

int
scm_iseq_gc_accept(ScmObj obj, ScmGCRefHandler handler)
{
  size_t idx, *offset;
  int rslt = SCM_GC_REF_HANDLER_VAL_INIT;

  scm_assert_obj_type(obj, &SCM_ISEQ_TYPE_INFO);

  EARY_FOR_EACH(SCM_ISEQ_EARY_OBJS(obj), idx, offset) {
    ScmRef chld
      = SCM_REF_MAKE_FROM_PTR((void *)(scm_iseq_to_ip(obj) + *offset));
    /* (void *) で cast しているのは Clang の -Wcast-align による警告を抑制するため */

    rslt = SCM_GC_CALL_REF_HANDLER(handler, obj, SCM_REF_DEREF(chld));
    if (scm_gc_ref_handler_failure_p(rslt))
      return rslt;
  }

  return rslt;
}
