#include <stdio.h>

#include "scythe/object.h"
#include "scythe/bedrock.h"
#include "scythe/vm.h"
#include "scythe/memory.h"
#include "scythe/refstk.h"
#include "scythe/exception.h"
#include "scythe/equivalence.h"
#include "scythe/format.h"
#include "scythe/iseq.h"
#include "scythe/miscobjects.h"
#include "scythe/pair.h"
#include "scythe/string.h"
#include "scythe/symbol.h"
#include "scythe/earray.h"
#include "scythe/procedure.h"


/*******************************************************************/
/*  Proc                                                           */
/*******************************************************************/

static void
scm_proc_set_env(ScmObj proc, ScmObj env)
{
  scm_assert(scm_obj_type_flag_set_p(proc, SCM_TYPE_FLG_PROC));

  SCM_WB_SETQ(proc, SCM_PROCEDURE(proc)->env, env);
}

ScmObj
scm_procedure_P(ScmObj proc)
{
  return (scm_procedure_p(proc) ? SCM_TRUE_OBJ : SCM_FALSE_OBJ);
}

int
scm_proc_initialize(ScmObj proc, ScmObj name, ScmProcApplyFunc apply,
                    int arity, unsigned int flags, ScmObj env)
{
  scm_assert(scm_obj_type_flag_set_p(proc, SCM_TYPE_FLG_PROC));
  scm_assert(apply != NULL);
  scm_assert(scm_obj_null_p(name) || scm_string_p(name) || scm_symbol_p(name));

  SCM_SLOT_SETQ(ScmProcedure, proc, name, name);
  SCM_SLOT_SETQ(ScmProcedure, proc, env, env);
  SCM_PROCEDURE(proc)->arity = arity;
  SCM_PROCEDURE(proc)->flags = flags;
  SCM_PROCEDURE(proc)->apply = apply;

  return 0;
}

void
scm_proc_gc_initialize(ScmObj obj)
{
  scm_assert(scm_obj_type_flag_set_p(obj, SCM_TYPE_FLG_PROC));

  SCM_PROCEDURE(obj)->name = SCM_OBJ_NULL;
  SCM_PROCEDURE(obj)->env = SCM_OBJ_NULL;
}

int
scm_proc_gc_accept(ScmObj obj, ScmGCRefHandler handler)
{
  int rslt = SCM_GC_REF_HANDLER_VAL_INIT;

  scm_assert(scm_obj_type_flag_set_p(obj, SCM_TYPE_FLG_PROC));

  rslt = SCM_GC_CALL_REF_HANDLER(handler, obj, SCM_PROCEDURE(obj)->name);
  if (scm_gc_ref_handler_failure_p(rslt)) return rslt;

  rslt = SCM_GC_CALL_REF_HANDLER(handler, obj, SCM_PROCEDURE(obj)->env);
  if (scm_gc_ref_handler_failure_p(rslt)) return rslt;

  return rslt;
}


/*******************************************************************/
/*  Subrutine                                                      */
/*******************************************************************/

ScmTypeInfo SCM_SUBRUTINE_TYPE_INFO = {
  .name                = "subrutine",
  .flags               = SCM_TYPE_FLG_PROC | SCM_TYPE_FLG_MMO,
  .obj_print_func      = scm_subrutine_obj_print,
  .obj_size            = sizeof(ScmSubrutine),
  .gc_ini_func         = scm_subrutine_gc_initialize,
  .gc_fin_func         = NULL,
  .gc_accept_func      = scm_subrutine_gc_accept,
  .gc_accept_func_weak = NULL,
  .extra               = NULL,
};

static int
subroutine_proc_apply_func(ScmObj proc, int argc, ScmObj vm)
{
  scm_assert_obj_type(proc, &SCM_SUBRUTINE_TYPE_INFO);

  return scm_vm_setup_to_call_subr(vm, proc, argc);
}

int
scm_subrutine_initialize(ScmObj subr, ScmSubrFunc func,
                         ScmObj name, int arity, unsigned int flags, ScmObj env)
{
  int rslt;

  SCM_REFSTK_INIT_REG(&name,
                      &subr, &env);

  scm_assert_obj_type(subr, &SCM_SUBRUTINE_TYPE_INFO);
  scm_assert(func != NULL);
  scm_assert(scm_obj_null_p(name) || scm_string_p(name) || scm_symbol_p(name));

  rslt = scm_proc_initialize(subr, name, subroutine_proc_apply_func,
                             arity, flags, env);
  if (rslt < 0) return -1;

  SCM_SUBRUTINE(subr)->subr_func = func;

  return 0;
}

ScmObj
scm_subrutine_new(scm_mem_type_t mtype,
                  ScmSubrFunc func, ScmObj name, int arity, unsigned int flags,
                  ScmObj env)
{
  ScmObj subr = SCM_OBJ_INIT;

  SCM_REFSTK_INIT_REG(&name,
                      &subr, &env);

  scm_assert(func != NULL);
  scm_assert(scm_obj_null_p(name) || scm_string_p(name) || scm_symbol_p(name));

  subr = scm_alloc_mem(&SCM_SUBRUTINE_TYPE_INFO, 0, mtype);
  if (scm_obj_null_p(subr)) return SCM_OBJ_NULL;

  if (scm_subrutine_initialize(subr, func, name, arity, flags, env))
    return SCM_OBJ_NULL;

  return subr;
}

int
scm_subrutine_obj_print(ScmObj obj, ScmObj port, int kind,
                        ScmObjPrintHandler handler)
{
  ScmObj name = SCM_OBJ_INIT;

  SCM_REFSTK_INIT_REG(&obj, &port,
                      &name);

  scm_assert_obj_type(obj, &SCM_SUBRUTINE_TYPE_INFO);

  name = scm_proc_name(obj);

  if (scm_obj_null_p(name))
    return scm_obj_default_print_func(obj, port, kind, handler);

  return scm_pformat_cstr(port,"#<subr ~a>", name, SCM_OBJ_NULL);
}

void
scm_subrutine_gc_initialize(ScmObj obj)
{
  scm_assert_obj_type(obj, &SCM_SUBRUTINE_TYPE_INFO);

  scm_proc_gc_initialize(obj);
}

int
scm_subrutine_gc_accept(ScmObj obj, ScmGCRefHandler handler)
{
  scm_assert_obj_type(obj, &SCM_SUBRUTINE_TYPE_INFO);

  return scm_proc_gc_accept(obj, handler);
}


/*******************************************************************/
/*  Closure                                                        */
/*******************************************************************/

ScmTypeInfo SCM_CLOSURE_TYPE_INFO = {
  .name                = "closure",
  .flags               = SCM_TYPE_FLG_PROC | SCM_TYPE_FLG_MMO,
  .obj_print_func      = NULL,
  .obj_size            = sizeof(ScmClosure),
  .gc_ini_func         = scm_closure_gc_initialize,
  .gc_fin_func         = NULL,
  .gc_accept_func      = scm_closure_gc_accept,
  .gc_accept_func_weak = NULL,
  .extra               = NULL,
};

static int
closure_proc_apply_func(ScmObj proc, int argc, ScmObj vm)
{
  scm_assert_obj_type(proc, &SCM_CLOSURE_TYPE_INFO);

  return scm_vm_setup_to_call_closure(vm, proc, argc, scm_closure_env(proc),
                                      scm_iseq_to_ip(scm_closure_iseq(proc)));
}

int
scm_closure_initialize(ScmObj clsr,
                       ScmObj iseq, ScmObj env, ScmObj name, int arity)
{
  int rslt;

  scm_assert_obj_type(clsr, &SCM_CLOSURE_TYPE_INFO);
  scm_assert(scm_iseq_p(iseq));
  scm_assert(scm_obj_null_p(name) || scm_string_p(name) || scm_symbol_p(name));

  rslt = scm_proc_initialize(clsr, name, closure_proc_apply_func,
                             arity, 0, env);
  if (rslt < 0) return -1;

  SCM_SLOT_SETQ(ScmClosure, clsr, iseq, iseq);

  return 0;
}

ScmObj
scm_closure_new(scm_mem_type_t mtype,
                ScmObj iseq, ScmObj env, ScmObj name, int arity)
{
  ScmObj clsr = SCM_OBJ_INIT;
  int rslt;

  SCM_REFSTK_INIT_REG(&iseq, &env, &name,
                      &clsr);

  scm_assert(scm_iseq_p(iseq));

  clsr = scm_alloc_mem(&SCM_CLOSURE_TYPE_INFO, 0, mtype);
  if (scm_obj_null_p(clsr)) return SCM_OBJ_NULL;

  rslt = scm_closure_initialize(clsr, iseq, env, name, arity);
  if (rslt < 0) return SCM_OBJ_NULL;

  return clsr;
}

scm_byte_t *
scm_closure_to_ip(ScmObj clsr)
{
  scm_assert(scm_closure_p(clsr));
  return scm_iseq_to_ip(scm_closure_iseq(clsr));
}

void
scm_closure_gc_initialize(ScmObj obj)
{
  scm_assert_obj_type(obj, &SCM_CLOSURE_TYPE_INFO);

  scm_proc_gc_initialize(obj);

  SCM_CLOSURE(obj)->iseq = SCM_OBJ_NULL;
}

int
scm_closure_gc_accept(ScmObj obj, ScmGCRefHandler handler)
{
  int rslt = SCM_GC_REF_HANDLER_VAL_INIT;

  scm_assert_obj_type(obj, &SCM_CLOSURE_TYPE_INFO);

  rslt = scm_proc_gc_accept(obj, handler);
  if (scm_gc_ref_handler_failure_p(rslt)) return rslt;

  rslt = SCM_GC_CALL_REF_HANDLER(handler, obj, SCM_CLOSURE(obj)->iseq);
  if (scm_gc_ref_handler_failure_p(rslt)) return rslt;

  return rslt;
}


/*******************************************************************/
/*  Continuation                                                   */
/*******************************************************************/

ScmTypeInfo SCM_DWHCALLERENV_TYPE_INFO = {
  .name                = "dwhcallerenv",
  .flags               = SCM_TYPE_FLG_MMO,
  .obj_print_func      = NULL,
  .obj_size            = sizeof(ScmDWHCallerEnv),
  .gc_ini_func         = scm_dwhcallerenv_gc_initialize,
  .gc_fin_func         = scm_dwhcallerenv_gc_finalize,
  .gc_accept_func      = scm_dwhcallerenv_gc_accept,
  .gc_accept_func_weak = NULL,
  .extra               = NULL,
};

int
scm_dwhcallerenv_initialize(ScmObj dwhce,
                            ScmObj cont, const ScmObj *val, size_t vc)
{
  scm_assert_obj_type(dwhce, &SCM_DWHCALLERENV_TYPE_INFO);
  scm_assert(scm_obj_not_null_p(cont));
  scm_assert(vc == 0 || val != NULL);

  SCM_SLOT_SETQ(ScmDWHCallerEnv, dwhce, cont, cont);

  if (vc == 0) {
    SCM_DWHCALLERENV(dwhce)->val = NULL;
  }
  else {
    SCM_DWHCALLERENV(dwhce)->val = scm_malloc(sizeof(ScmObj) * vc);
    if (SCM_DWHCALLERENV(dwhce)->val == NULL)
      return -1;
    SCM_DWHCALLERENV(dwhce)->vc = vc;
  }

  for (size_t i = 0; i < vc; i++)
    SCM_SLOT_SETQ(ScmDWHCallerEnv, dwhce, val[i], val[i]);

  return 0;
}

void
scm_dwhcallerenv_finalize(ScmObj dwhce)
{
  scm_assert_obj_type(dwhce, &SCM_DWHCALLERENV_TYPE_INFO);

  if (SCM_DWHCALLERENV(dwhce)->vc == 0)
    return ;

  scm_free(SCM_DWHCALLERENV(dwhce)->val);
  SCM_DWHCALLERENV(dwhce)->val = NULL;
  SCM_DWHCALLERENV(dwhce)->vc = 0;
}

ScmObj
scm_dwhcallerenv_new(scm_mem_type_t mtype,
                     ScmObj cont, const ScmObj *val, size_t vc)
{
  ScmObj dwhce = SCM_OBJ_INIT;

  SCM_REFSTK_INIT_REG(&cont,
                      &dwhce);

  scm_assert(scm_obj_not_null_p(cont));
  scm_assert(vc == 0 || val != NULL);

  dwhce = scm_alloc_mem(&SCM_DWHCALLERENV_TYPE_INFO, 0, mtype);
  if (scm_obj_null_p(dwhce)) return SCM_OBJ_NULL;

  if (scm_dwhcallerenv_initialize(dwhce, cont, val, vc) < 0)
    return SCM_OBJ_NULL;

  return dwhce;
}

void
scm_dwhcallerenv_gc_initialize(ScmObj obj)
{
  scm_assert_obj_type(obj, &SCM_DWHCALLERENV_TYPE_INFO);

  SCM_DWHCALLERENV(obj)->cont = SCM_OBJ_NULL;
  SCM_DWHCALLERENV(obj)->vc = 0;
}

void
scm_dwhcallerenv_gc_finalize(ScmObj obj)
{
  scm_dwhcallerenv_finalize(obj);
}

int
scm_dwhcallerenv_gc_accept(ScmObj obj, ScmGCRefHandler handler)
{
  int rslt = SCM_GC_REF_HANDLER_VAL_INIT;

  rslt = SCM_GC_CALL_REF_HANDLER(handler,
                                 obj, SCM_DWHCALLERENV(obj)->cont);
  if (scm_gc_ref_handler_failure_p(rslt)) return rslt;

  for (size_t i = 0; i < SCM_DWHCALLERENV(obj)->vc; i++) {
    rslt = SCM_GC_CALL_REF_HANDLER(handler,
                                   obj, SCM_DWHCALLERENV(obj)->val[i]);
    if (scm_gc_ref_handler_failure_p(rslt)) return rslt;
  }

  return rslt;
}

static int
call_dw_handler(ScmObj caller, ScmObj handlers)
{
  scm_assert(scm_procedure_p(caller));
  scm_assert(scm_pair_p(handlers));

  return scm_trampolining(scm_car(handlers), SCM_NIL_OBJ,
                          caller, scm_cdr(handlers));
}

static int
scm_cont_subr_func_dw_handler_calller(ScmObj subr, int argc, const ScmObj *argv)
{
  ScmObj env = SCM_OBJ_INIT;
  int r;

  SCM_REFSTK_INIT_REG(&subr,
                      &env);

  if (!scm_pair_p(argv[0])) {
    env = scm_proc_env(subr);
    scm_assert_obj_type(env, &SCM_DWHCALLERENV_TYPE_INFO);

    r = scm_reinstantemnet_continuation(scm_dwhcallerenv_cont(env));
    if (r < 0) return -1;

    return scm_return_val(scm_dwhcallerenv_val(env),
                          (int)scm_dwhcallerenv_vc(env));
  }
  else {
    return call_dw_handler(subr, argv[0]);
  }
}

int
scm_cont_subr_func_continuation(ScmObj subr, int argc, const ScmObj *argv)
{
  ScmObj handlers = SCM_OBJ_INIT, caller = SCM_OBJ_INIT, env = SCM_OBJ_INIT;

  SCM_REFSTK_INIT_REG(&subr,
                      &handlers, &caller);

  scm_assert(scm_continuation_p(subr));
  scm_assert(argc >= 0);

  handlers = scm_collect_dynamic_wind_handler(scm_proc_env(subr));
  if (scm_obj_null_p(handlers)) return -1;

  if (!scm_pair_p(handlers)) {
    int r = scm_reinstantemnet_continuation(scm_proc_env(subr));
    if (r < 0) return -1;

    return scm_return_val(argv, argc);
  }
  else {
    env = scm_dwhcallerenv_new(SCM_MEM_HEAP,
                               scm_proc_env(subr), argv, (size_t)argc);
    if (scm_obj_null_p(env)) return -1;

    caller = scm_make_subrutine(scm_cont_subr_func_dw_handler_calller,
                                -2, SCM_PROC_ADJ_UNWISHED, env);
    if (scm_obj_null_p(caller)) return -1;

    return call_dw_handler(caller, handlers);
  }
}

ScmObj
scm_continuation_new(scm_mem_type_t mtype, ScmObj contcap)
{
  ScmObj name = SCM_OBJ_INIT;

  SCM_REFSTK_INIT_REG(&contcap,
                      &name);

  name = scm_make_symbol_from_cstr("continuation", SCM_ENC_SRC);
  if (scm_obj_null_p(name)) return SCM_OBJ_NULL;

  return scm_subrutine_new(mtype, scm_cont_subr_func_continuation,
                           name, -1, SCM_PROC_ADJ_UNWISHED, contcap);
}

ScmObj
scm_make_continuation(void)
{
  ScmObj cap = SCM_OBJ_INIT;

  SCM_REFSTK_INIT_REG(&cap);

  cap = scm_capture_continuation();
  if (scm_obj_null_p(cap)) return SCM_OBJ_NULL;

  return scm_continuation_new(SCM_MEM_HEAP, cap);
}


/*******************************************************************/
/*  Parameter                                                      */
/*******************************************************************/

int
scm_prm_subr_func_parameter__init(ScmObj subr, int argc, const ScmObj *argv)
{
  scm_parameter_set_init_val(argv[0], argv[1]);
  return scm_return_val(argv, 1);
}

int
scm_prm_subr_func_parameter__cons(ScmObj subr, int argc, const ScmObj *argv)
{
  return scm_return_val_1(scm_cons(argv[0], argv[1]));
}

/*
 * 引数無しでの parameter オブジェクトの呼び出し
 *   (<parameter object>)
 *      ;=> parameter オブジェクトを(動的に)束縛している値を返す
 *
 * 引数を 1 つ指定した parameter オブジェクトの呼び出し
 *   (<parameter object> arg1)
 *      ;=> 何もせずただ #<undef> を返す
 *
 * 引数を 2 つ指定し、第 2 引数が #t である場合の parameter オブジェクトの呼び
 * 出し
 *   (<parameter object> arg1 #t)
 *      ;=> (<converter> arg1) の結果を parameter オブジェクの初期値に設定し、
 *          parameter オブジェクトを返す
 *
 * 引数を 2 つ指定し、第 2 引数が #f である場合の parameter オブジェクトの呼び
 * 出し
 *   (<parameter object> arg1 #f)
 *      ;=> (<parameter object> . {(<converter> arg1) の結果}) を返す
 *
 * 引数を 2 つ指定し、第 2 引数が真偽値でない場合の parameter オブジェクトの呼
 * び出し
 *   (<parameter object> arg1 arg2)
 *      ;=> 何もせずただ #<undef> を返す
 *
 */
int
scm_prm_subr_func_parameter(ScmObj subr, int argc, const ScmObj *argv)
{
  ScmObj val = SCM_OBJ_INIT, conv = SCM_OBJ_INIT, postproc = SCM_OBJ_INIT;
  ScmObj arg1 = SCM_OBJ_INIT, arg2 = SCM_OBJ_INIT;

  SCM_REFSTK_INIT_REG(&subr,
                      &val, &conv, &postproc,
                      &arg1, &arg2);

  scm_assert(scm_parameter_p(subr));

  if (scm_nil_p(argv[0]))
    return scm_return_val_1(scm_parameter_value(subr));

  arg1 = scm_car(argv[0]);
  if (scm_nil_p(scm_cdr(argv[0])))
    arg2 = SCM_OBJ_NULL;
  else
    arg2 = scm_list_ref(argv[0], 1);

  if (scm_obj_null_p(arg2)) {
    return scm_return_val_1(SCM_UNDEF_OBJ);
  }
  else if (scm_true_object_p(arg2)) {
    conv = scm_parameter_converter(subr);
    if (scm_procedure_p(conv)) {
      arg1 = scm_cons(arg1, SCM_NIL_OBJ);
      if (scm_obj_null_p(arg1)) return -1;

      postproc = scm_premade_procedure(SCM_PREMADE_PROC_PARAMETER__INIT);
      return scm_trampolining(conv, arg1, postproc, subr);
    }
    else {
      scm_parameter_set_init_val(subr, arg1);
      return scm_return_val_1(subr);
    }
  }
  else if (scm_false_object_p(arg2)){
    conv = scm_parameter_converter(subr);
    if (scm_procedure_p(conv)) {
      arg1 = scm_cons(arg1, SCM_NIL_OBJ);
      if (scm_obj_null_p(arg1)) return -1;

      postproc = scm_premade_procedure(SCM_PREMADE_PROC_PARAMETER__CONS);
      return scm_trampolining(conv, arg1, postproc, subr);
    }
    else {
      return scm_return_val_1(scm_cons(subr, arg1));
    }
  }
  else {
    return scm_return_val_1(SCM_UNDEF_OBJ);
  }
}

ScmObj
scm_parameter_new(scm_mem_type_t mtype, ScmObj init, ScmObj conv)
{
  ScmObj name = SCM_OBJ_INIT, env = SCM_OBJ_INIT;

  SCM_REFSTK_INIT_REG(&init, &conv,
                      &name, &env);

  scm_assert(scm_obj_null_p(conv) || scm_procedure_p(conv));

  name = scm_make_symbol_from_cstr("parameter", SCM_ENC_SRC);
  if (scm_obj_null_p(name)) return SCM_OBJ_NULL;

  env = scm_cons(scm_obj_null_p(init) ? SCM_UNDEF_OBJ : init,
                 scm_obj_null_p(conv) ? SCM_UNDEF_OBJ : conv);
  if (scm_obj_null_p(env)) return SCM_OBJ_NULL;

  return scm_subrutine_new(mtype, scm_prm_subr_func_parameter,
                           name, -1, 0, env);
}

ScmObj
scm_parameter_init_val(ScmObj prm)
{
  scm_assert(scm_parameter_p(prm));
  return scm_car(scm_proc_env(prm));
}

ScmObj
scm_parameter_converter(ScmObj prm)
{
  scm_assert(scm_parameter_p(prm));
  return scm_cdr(scm_proc_env(prm));
}

void
scm_parameter_set_init_val(ScmObj prm, ScmObj val)
{
  scm_assert(scm_parameter_p(prm));
  scm_assert(scm_obj_not_null_p(val));
  scm_set_car(scm_proc_env(prm), val);
}


/*******************************************************************/
/*  CdProgram                                                      */
/*******************************************************************/

ScmTypeInfo SCM_CDPROGRAM_TYPE_INFO = {
  .name                = "cdprogram",
  .flags               = SCM_TYPE_FLG_PROC | SCM_TYPE_FLG_MMO,
  .obj_print_func      = scm_cdprogram_obj_print,
  .obj_size            = sizeof(ScmCdProgram),
  .gc_ini_func         = scm_cdprogram_gc_initialize,
  .gc_fin_func         = scm_cdprogram_gc_finalize,
  .gc_accept_func      = scm_cdprogram_gc_accept,
  .gc_accept_func_weak = NULL,
  .extra               = NULL,
};

static int
cdprogram_proc_apply_func(ScmObj proc, int argc, ScmObj vm)
{
  scm_assert_obj_type(proc, &SCM_CDPROGRAM_TYPE_INFO);

  return scm_vm_setup_to_call_closure(vm, proc, argc, SCM_OBJ_NULL,
                                      scm_iseq_to_ip(scm_cdprogram_head_iseq(proc)));
}

int
scm_cdprogram_initialize(ScmObj prog, ScmObj name, ScmObj iseq)
{
  int r;

  SCM_REFSTK_INIT_REG(&prog, &name, &iseq);

  scm_assert_obj_type(prog, &SCM_CDPROGRAM_TYPE_INFO);
  scm_assert(scm_obj_null_p(name) || scm_string_p(name) || scm_symbol_p(name));
  scm_assert(scm_obj_null_p(iseq) || scm_iseq_p(iseq));

  if (scm_obj_null_p(name)) {
    name = scm_make_symbol_from_cstr("anonymous", SCM_ENC_SRC);
    if (scm_obj_null_p(name)) return -1;
  }

  r = scm_proc_initialize(prog, name, cdprogram_proc_apply_func,
                          0, 0, SCM_OBJ_NULL);
  if (r < 0) return -1;

  if (scm_obj_null_p(iseq)) {
    iseq = scm_make_iseq();
    if (scm_obj_null_p(iseq)) return -1;
  }

  r = eary_init(SCM_CDPROGRAM_SEQS(prog), sizeof(ScmObj), 1);
  if (r < 0) return -1;

  EARY_PUSH_SCMOBJ(SCM_CDPROGRAM_SEQS(prog), iseq, prog, r);
  if (r < 0) return -1;

  return 0;
}

ScmObj
scm_cdprogram_new(scm_mem_type_t mtype, ScmObj name, ScmObj iseq)
{
  ScmObj prog = SCM_OBJ_INIT;

  SCM_REFSTK_INIT_REG(&prog, &name, &iseq);

  scm_assert(scm_obj_null_p(name) || scm_string_p(name) || scm_symbol_p(name));
  scm_assert(scm_obj_null_p(iseq) || scm_iseq_p(iseq));

  prog = scm_alloc_mem(&SCM_CDPROGRAM_TYPE_INFO, 0, mtype);
  if (scm_obj_null_p(prog)) return SCM_OBJ_NULL;

  if (scm_cdprogram_initialize(prog, name, iseq) < 0)
    return SCM_OBJ_NULL;

  return prog;
}

void
scm_cdprogram_finalize(ScmObj obj)
{
  scm_assert_obj_type(obj, &SCM_CDPROGRAM_TYPE_INFO);

  eary_fin(SCM_CDPROGRAM_SEQS(obj));
}

ssize_t
scm_cdprogram_assign_id(ScmObj prog, ssize_t id)
{
  size_t i, j, s;
  int r;

  scm_assert_obj_type(prog, &SCM_CDPROGRAM_TYPE_INFO);

  i = (id >= 0) ? (size_t)id : EARY_SIZE(SCM_CDPROGRAM_SEQS(prog));
  if (i >= SSIZE_MAX) {
    scm_error("failed to assign assembler block id: too many blocks", 0);
    return -1;
  }

  s = EARY_SIZE(SCM_CDPROGRAM_SEQS(prog));
  for (j = i; j > s; j--) {
    EARY_SET_SCMOBJ(SCM_CDPROGRAM_SEQS(prog), j, SCM_OBJ_NULL, prog, r);
    if (r < 0) return -1;
  }

  if (j == s) {
    EARY_SET_SCMOBJ(SCM_CDPROGRAM_SEQS(prog), j, SCM_OBJ_NULL, prog, r);
    if (r < 0) return -1;
  }

  return (ssize_t)i;
}

bool
scm_cdprogram_iseq_exist_p(ScmObj prog, size_t id)
{
  ScmObj iseq = SCM_OBJ_INIT;

  scm_assert(scm_cdprogram_p(prog));

  if (!scm_cdprogram_assigned_id_p(prog, id))
    return false;

  EARY_GET(SCM_CDPROGRAM_SEQS(prog), ScmObj, id, iseq);
  return scm_obj_not_null_p(iseq);
}

static ScmObj
cdprogram_create_iseq(ScmObj prog, size_t id)
{
  ScmObj iseq = SCM_OBJ_INIT;
  int r;

  SCM_REFSTK_INIT_REG(&prog,
                      &iseq);

  scm_assert_obj_type(prog, &SCM_CDPROGRAM_TYPE_INFO);

  iseq = scm_make_iseq();
  if (scm_obj_null_p(iseq)) return SCM_OBJ_NULL;

  EARY_SET_SCMOBJ(SCM_CDPROGRAM_SEQS(prog), id, iseq, prog, r);
  if (r < 0) return SCM_OBJ_NULL;

  return iseq;
}

ScmObj
scm_cdprogram_create_iseq(ScmObj prog, size_t id)
{
  scm_assert_obj_type(prog, &SCM_CDPROGRAM_TYPE_INFO);

  if (id >= EARY_SIZE(SCM_CDPROGRAM_SEQS(prog))) {
    scm_error("failed to create iseq: "
              "specified sequence id has not assigned", 0);
    return SCM_OBJ_NULL;
  }

  if (scm_cdprogram_iseq_exist_p(prog, id)) {
    scm_error("failed to create iseq: "
              "iseq for specified sequence id has already been created", 0);
    return SCM_OBJ_NULL;
  }

  return cdprogram_create_iseq(prog, id);
}

ScmObj
scm_cdprogram_iseq(ScmObj prog, size_t id, bool create)
{
  ScmObj iseq = SCM_OBJ_INIT;

  scm_assert_obj_type(prog, &SCM_CDPROGRAM_TYPE_INFO);

  if (id >= EARY_SIZE(SCM_CDPROGRAM_SEQS(prog))) {
    scm_error("failed to refer iseq: "
              "specified sequence id has not assigned", 0);
    return SCM_OBJ_NULL;
  }

  EARY_GET(SCM_CDPROGRAM_SEQS(prog), ScmObj, id, iseq);
  if (scm_obj_not_null_p(iseq))
    return iseq;

  if (create)
    return cdprogram_create_iseq(prog, id);

  scm_error("failed to refer iseq: "
            "iseq for specified sequence id is not created", 0);
  return SCM_OBJ_NULL;
}

ScmObj
scm_cdprogram_next_iseq(ScmObj prog, size_t *id)
{
  scm_assert_obj_type(prog, &SCM_CDPROGRAM_TYPE_INFO);
  scm_assert(id != NULL);

  if (*id >= EARY_SIZE(SCM_CDPROGRAM_SEQS(prog)) - 1)
    return SCM_OBJ_NULL;

 for (size_t i = *id + 1; i < EARY_SIZE(SCM_CDPROGRAM_SEQS(prog)); i++) {
   ScmObj iseq;
   EARY_GET(SCM_CDPROGRAM_SEQS(prog), ScmObj, i, iseq);
   if (scm_obj_not_null_p(iseq)) {
     *id = i;
     return iseq;
   }
 }

 return SCM_OBJ_NULL;
}

ssize_t
scm_cdprogram_id(ScmObj prog, ScmObj iseq)
{
  ScmObj *itr = SCM_OBJ_INIT;
  size_t idx;

  scm_assert_obj_type(prog, &SCM_CDPROGRAM_TYPE_INFO);
  scm_assert(scm_iseq_p(iseq));

  EARY_FOR_EACH(SCM_CDPROGRAM_SEQS(prog), idx, itr) {
    if (scm_eq_p(*itr, iseq))
      return (ssize_t)idx;
  }

  return -1;
}

void
scm_cdprogram_set_log(ScmObj prog, ScmObj log)
{
  scm_assert(scm_cdprogram_p(prog));

  return scm_proc_set_env(prog, log);
}

scm_byte_t *
scm_cdprogram_to_ip(ScmObj prog)
{
  scm_assert(scm_cdprogram_p(prog));

  return scm_iseq_to_ip(scm_cdprogram_head_iseq(prog));
}

int
scm_cdprogram_obj_print(ScmObj obj, ScmObj port, int kind,
                        ScmObjPrintHandler handler)
{
  ScmObj name = SCM_OBJ_INIT;

  SCM_REFSTK_INIT_REG(&obj, &port,
                      &name);

  scm_assert_obj_type(obj, &SCM_SUBRUTINE_TYPE_INFO);

  name = scm_proc_name(obj);

  if (scm_obj_null_p(name))
    return scm_obj_default_print_func(obj, port, kind, handler);

  return scm_pformat_cstr(port,"#<compiled-program ~a>", name, SCM_OBJ_NULL);
}

void
scm_cdprogram_gc_initialize(ScmObj obj)
{
  scm_assert_obj_type(obj, &SCM_CDPROGRAM_TYPE_INFO);

  scm_proc_gc_initialize(obj);

  eary_init(SCM_CDPROGRAM_SEQS(obj), 0, 0);
}

void
scm_cdprogram_gc_finalize(ScmObj obj)
{
  scm_cdprogram_finalize(obj);
}

int
scm_cdprogram_gc_accept(ScmObj obj, ScmGCRefHandler handler)
{
  int rslt = SCM_GC_REF_HANDLER_VAL_INIT;
  ScmObj *iseq;
  size_t idx;

  scm_assert_obj_type(obj, &SCM_CDPROGRAM_TYPE_INFO);

  rslt = scm_proc_gc_accept(obj, handler);
  if (scm_gc_ref_handler_failure_p(rslt)) return rslt;

  EARY_FOR_EACH(SCM_CDPROGRAM_SEQS(obj), idx, iseq) {
    rslt = SCM_GC_CALL_REF_HANDLER(handler, obj, *iseq);
    if (scm_gc_ref_handler_failure_p(rslt)) return rslt;
  }

  return rslt;
}


/* XXX: Marshal/Unmarshal の動作確認に使用する目的で作成した等価性評価関数な
 *      ので、テスト目的以外には使用しない
 */

#include "scythe/assembler.h"
#include "scythe/compiler.h"
#include "scythe/module.h"

int
scm_cdprogram_eq(ScmObj prog1, ScmObj prog2, bool *rslt)
{
  ScmObj dasm1 = SCM_OBJ_INIT, dasm2 = SCM_OBJ_INIT;
  ScmObj name1 = SCM_OBJ_NULL, name2 = SCM_OBJ_INIT;
  int r;

  SCM_REFSTK_INIT_REG(&prog1, &prog2,
                      &dasm1, &dasm2,
                      &name1, &name2);

  scm_assert_obj_type(prog1, &SCM_CDPROGRAM_TYPE_INFO);
  scm_assert_obj_type(prog2, &SCM_CDPROGRAM_TYPE_INFO);

  r = scm_equal(scm_proc_name(prog1), scm_proc_name(prog2), rslt);
  if (r < 0) return -1;

  if (!*rslt)
    return 0;

  dasm1 = scm_make_disassembler(prog1);
  if (scm_obj_null_p(dasm1)) return -1;

  dasm2 = scm_make_disassembler(prog2);
  if (scm_obj_null_p(dasm2)) return -1;

  while (!scm_disasm_terminate_p(dasm1)) {
    if (scm_disasm_fmt(dasm1) != scm_disasm_fmt(dasm2))
      goto not_equal;

    if (scm_disasm_op(dasm1) != scm_disasm_op(dasm2))
      goto not_equal;

    switch (scm_disasm_fmt(dasm1)) {
    case SCM_ASM_FMT_NOOPD:
      break;
    case SCM_ASM_FMT_OBJ:
      if (scm_qqtmpl_p(scm_disasm_opd_obj__1(dasm1))
          && scm_qqtmpl_p(scm_disasm_opd_obj__1(dasm2)))
        r = scm_qqtmpl_eq(scm_disasm_opd_obj__1(dasm1),
                          scm_disasm_opd_obj__1(dasm2),
                          rslt);
      else
        r = scm_equal(scm_disasm_opd_obj__1(dasm1),
                      scm_disasm_opd_obj__1(dasm2),
                      rslt);
      if (r < 0) return -1;
      if (!*rslt) return 0;
      break;
    case SCM_ASM_FMT_GV_MOD:
      r = scm_equal(scm_disasm_opd_gv_mod__1(dasm1),
                    scm_disasm_opd_gv_mod__1(dasm2),
                    rslt);
      if (r < 0) return -1;
      if (!*rslt) return 0;
      if (scm_disasm_opd_gv_mod__mod_id(dasm1) >= 0) {
        if (scm_disasm_opd_gv_mod__mod_id(dasm1)
            != scm_disasm_opd_gv_mod__mod_id(dasm2))
          goto not_equal;
      }
      else {
        r = scm_equal(scm_disasm_opd_gv_mod__2(dasm1),
                      scm_disasm_opd_gv_mod__2(dasm2),
                      rslt);
        if (r < 0) return -1;
        if (!*rslt) return 0;
      }
      break;
    case SCM_ASM_FMT_SI:
      if (scm_disasm_opd_si__1(dasm1) != scm_disasm_opd_si__1(dasm2))
        goto not_equal;
      break;
    case SCM_ASM_FMT_SI_SI:
      if (scm_disasm_opd_si_si__1(dasm1) != scm_disasm_opd_si_si__1(dasm2)
          || scm_disasm_opd_si_si__2(dasm1) != scm_disasm_opd_si_si__2(dasm2))
        goto not_equal;
      break;
    case SCM_ASM_FMT_SI_SI_SEQ:
      if (scm_disasm_opd_si_si_seq__1(dasm1) != scm_disasm_opd_si_si_seq__1(dasm2)
          || scm_disasm_opd_si_si_seq__2(dasm1) != scm_disasm_opd_si_si_seq__2(dasm2)
          || scm_disasm_opd_si_si_seq__seq_id(dasm1) != scm_disasm_opd_si_si_seq__seq_id(dasm2))
        goto not_equal;
      break;
    case SCM_ASM_FMT_IOF:
      if (scm_disasm_opd_iof__1(dasm1) != scm_disasm_opd_iof__1(dasm2))
        goto not_equal;
      break;
    case SCM_ASM_FMT_MOD:
      if (scm_disasm_opd_mod__mod_id(dasm1) >= 0) {
        if (scm_disasm_opd_mod__mod_id(dasm1) != scm_disasm_opd_mod__mod_id(dasm2))
          goto not_equal;
      }
      else {
        r = scm_equal(scm_disasm_opd_mod__1(dasm1),
                      scm_disasm_opd_mod__1(dasm2),
                      rslt);
        if (r < 0) return -1;
        if (!*rslt) return 0;
      }
      break;
    case SCM_ASM_PI_LABEL:
      if (scm_disasm_opd_label__id(dasm1) != scm_disasm_opd_label__id(dasm2))
        goto not_equal;
      break;
    case SCM_ASM_PI_DECLMOD:
      if (scm_disasm_opd_declmod__id(dasm1) != scm_disasm_opd_declmod__id(dasm2))
        goto not_equal;
      name1 = scm_module_name(scm_disasm_opd_declmod__mod(dasm1));
      name2 = scm_module_name(scm_disasm_opd_declmod__mod(dasm2));
      r = scm_equal(name1, name2, rslt);
      if (r < 0) return -1;
      if (!*rslt) return 0;
      if (scm_disasm_opd_declmod__reg(dasm1) != scm_disasm_opd_declmod__reg(dasm2))
        goto not_equal;
      break;
    case SCM_ASM_PI_SEQUENCE:
      if (scm_disasm_opd_sequence__id(dasm1) != scm_disasm_opd_sequence__id(dasm2))
        goto not_equal;
      break;
    default:
      scm_error("failed to compare <cdprogram> objects", 0);
      return -1;
    }

    r = scm_disasm_next(dasm1);
    if (r < 0) return -1;

    r = scm_disasm_next(dasm2);
    if (r < 0) return -1;
  }

  if (!scm_disasm_terminate_p(dasm2))
    goto not_equal;

  if (rslt != NULL) *rslt = true;
  return 0;

 not_equal:
  if (rslt != NULL) *rslt = false;
  return 0;
}
